"""coBib's BibTex parser.

This parser leverages the [`bibtexparser`](https://pypi.org/project/bibtexparser/) library to
convert between `cobib.database.Entry` instances and raw BibTex strings.

Non-standard BibTex types can be configured to be ignored via
`cobib.config.config.BibtexParserConfig.ignore_non_standard_types`.

The parser is registered under the `-b` and `--bibtex` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import logging
from collections import OrderedDict
from typing import Dict

import bibtexparser
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)


class BibtexParser(Parser):
    """The BibTex Parser."""

    name = "bibtex"

    @override
    def parse(self, string: str) -> Dict[str, Entry]:
        string = Event.PreBibtexParse.fire(string) or string

        bparser = bibtexparser.bparser.BibTexParser()
        bparser.ignore_nonstandard_types = config.parsers.bibtex.ignore_non_standard_types
        bparser.common_strings = True
        bparser.interpolate_strings = False
        try:
            LOGGER.debug("Loading BibTex data from file: %s.", string)
            with open(string, "r", encoding="utf-8") as file:
                database = bibtexparser.load(file, parser=bparser)
        except (OSError, FileNotFoundError):
            LOGGER.debug("Loading BibTex string: %s.", string)
            database = bibtexparser.loads(string, parser=bparser)
        bib = OrderedDict()
        for entry in database.entries:
            if "month" in entry.keys() and isinstance(
                entry["month"], bibtexparser.bibtexexpression.BibDataStringExpression
            ):
                entry["month"] = entry["month"].expr[0].name
            label = entry.pop("ID")
            bib[label] = Entry(label, entry)

        Event.PostBibtexParse.fire(bib)

        return bib

    @override
    def dump(self, entry: Entry) -> str:
        Event.PreBibtexDump.fire(entry)

        database = bibtexparser.bibdatabase.BibDatabase()
        stringified_entry = entry.stringify()
        stringified_entry["ID"] = stringified_entry.pop("label")
        if "month" in stringified_entry.keys():
            # convert month to bibtexexpression
            stringified_entry["month"] = bibtexparser.bibtexexpression.BibDataStringExpression(
                [bibtexparser.bibdatabase.BibDataString(database, stringified_entry["month"])]
            )
        database.entries = [stringified_entry]
        LOGGER.debug("Converting entry %s to BibTex format.", entry.label)
        writer = bibtexparser.bwriter.BibTexWriter()
        writer.common_strings = True
        string: str = writer.write(database)

        string = Event.PostBibtexDump.fire(string) or string

        return string



"""coBib's Show command.

This command simply shows/prints the specified entry as a BibLaTex-formmatted string.
```
cobib show <label>
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `ENTER` key.
"""

from __future__ import annotations

import logging
from typing import List, Type

from rich.console import Console, ConsoleRenderable
from rich.prompt import PromptBase, PromptType
from rich.syntax import Syntax
from textual.app import App
from typing_extensions import override

from cobib import __version__
from cobib.config import Event
from cobib.database import Database
from cobib.parsers.bibtex import BibtexParser

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class ShowCommand(Command):
    """The Show Command.

    This command can parse the following arguments:

        * `label`: the label of the entry to be shown.
    """

    name = "show"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.entry_str: str = ""
        """The string-formatted `cobib.database.Entry` shown by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="show", description="Show subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Show command.")

        Event.PreShowCommand.fire(self)

        try:
            entry = Database()[self.largs.label]
            entry_str = BibtexParser().dump(entry)

            self.entry_str = entry_str
        except KeyError:
            msg = f"No entry with the label '{self.largs.label}' could be found."
            LOGGER.error(msg)

        Event.PostShowCommand.fire(self)

    @override
    def render_porcelain(self) -> List[str]:
        return self.entry_str.split("\n")

    @override
    def render_rich(self, *, background_color: str = "default") -> ConsoleRenderable:
        syntax = Syntax(self.entry_str, "bibtex", background_color=background_color, word_wrap=True)
        return syntax



"""coBib's input screen.

This screen is an interactive input screen. It automatically focuses the central `Input` widget but
also provides panels to displays user prompts, help information and other popups.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Input, Static
from typing_extensions import override

from .popup_panel import PopupPanel


class InputScreen(ModalScreen[str]):
    """coBib's input screen."""

    AUTO_FOCUS = "Input"

    BINDINGS = [("escape", "escape", "Quit the prompt")]
    """
    | Key(s) | Description |
    | :- | :- |
    | escape | When enabled, quits the interactive input screen. |
    """

    DEFAULT_CSS = """
        InputScreen {
            align: center middle;
        }

        #input {
            padding: 1 2;
            width: 80%;
            height: auto;
            background: $surface;
        }

        #prompt {
            padding: 1 2;
            width: 80%;
            height: auto;
            background: $surface;
        }

        #help {
            padding: 1 2;
            width: 80%;
            height: auto;
            background: $surface;
        }

        #error {
            padding: 1 2;
            width: 80%;
            height: auto;
            background: red 20%;
        }

        #panel {
            layer: default;
            layout: vertical;
            align: center bottom;
            padding: 0 2;
            width: 80%;
            height: auto;
            background: $surface;
        }
    """

    escape_enabled: bool = True
    """Enables the `escape` key binding."""

    @override
    def compose(self) -> ComposeResult:
        yield PopupPanel(id="panel")
        inp = Input(id="input")
        yield inp

    async def on_resume(self) -> None:
        """The hook triggered when receiving the `Resume` event.

        Basically this method clears all pre-existing popups.
        """
        await self.query_one(PopupPanel).query(Widget).remove()

    async def action_escape(self) -> None:
        """The action to perform when the `Escape` key is pressed."""
        if self.escape_enabled:
            self.dismiss("")
            await self.query(Static).remove()

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        """The action to perform when receiving the `Submitted` event.

        In this case, unmount the widget itself if `catch` is set. Otherwise the event gets
        [bubbled up](https://textual.textualize.io/guide/events/#bubbling).
        """
        self.dismiss(event.input.value)
        await self.query(Static).remove()
        event.stop()



"""coBib's motion key event.

This custom event gets send by coBib's widgets and is used to handle cursor motions.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from textual.events import Event


class MotionKey(Event):
    """coBib's motion key event."""

    def __init__(self, key: str) -> None:
        """Initializes a motion key event.

        Args:
            key: the motion key which was pressed to trigger this event.
        """
        super().__init__()
        self.key = key



"""coBib's Init command.

This command must be the first command you ever execute with coBib because it takes care of
initializing the database.
Generally, you will only ever need to run this command once, but it is safe to run it multiple times
(although it will likely have no effect).

To get started with coBib you must run:
```
cobib init
```
This will initialize the database in the location specified by
`cobib.config.config.DatabaseConfig.file`.

If you enabled the automatic git-integration of coBib via `cobib.config.config.DatabaseConfig.git`,
you must initialize this separately via:
```
cobib init --git
```
If you have not run the first command yet, you can directly initialize the database *and* the
git-integration by only running the second command.

.. warning::
   You can**not** run this command from the TUI, because the database must have already been
   initialized *before* you can start the TUI in the first place.
"""

from __future__ import annotations

import logging
import os
import sys
from pathlib import Path
from typing import Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class InitCommand(Command):
    """The Init Command.

    This command can parse the following arguments:

        * `-g`, `--git`: initializes the git-integration.
    """

    name = "init"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.file: Path
        """The path to the database file."""

        self.root: Path
        """The parent directory where the database file resides. This is where the git repository
        gets initialized (if the git integration was enabled)."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="init", description="Init subcommand parser.")
        parser.add_argument("-g", "--git", action="store_true", help="initialize git repository")
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Init command.")

        Event.PreInitCommand.fire(self)

        self.file = RelPath(config.database.file).path
        self.root = self.file.parent

        file_exists = self.file.exists()
        git_tracked = (self.root / ".git").exists()

        if file_exists:
            if git_tracked:
                msg = (
                    "Database file already exists and is being tracked by git. There is nothing "
                    "else to do."
                )
                LOGGER.info(msg)
                return

            if not git_tracked and not self.largs.git:
                msg = "Database file already exists! Use --git to start tracking it with git."
                LOGGER.warning(msg)
                return

        else:
            LOGGER.debug('Creating path for database file: "%s"', self.root)
            self.root.mkdir(parents=True, exist_ok=True)

            LOGGER.debug('Creating empty database file: "%s"', self.file)
            open(self.file, "w", encoding="utf-8").close()  # pylint: disable=consider-using-with

        if self.largs.git:
            if not config.database.git:
                msg = (
                    "You are about to initialize the git tracking of your database, but this will "
                    "only have effect if you also enable the DATABASE/git setting in your "
                    "configuration file!"
                )
                LOGGER.warning(msg)
            # First, check whether git is configured correctly.
            print("Checking `git config --get user.name`:", end=" ", flush=True)
            name_set = os.system("git config --get user.name")
            print()
            print("Checking `git config --get user.email`:", end=" ", flush=True)
            email_set = os.system("git config --get user.email")
            print()
            if name_set != 0 or email_set != 0:
                msg = (
                    "In order to use git you must configure your name and email first! For more "
                    "information please consult `man gittutorial`."
                )
                LOGGER.warning(msg)
                sys.exit(1)
            LOGGER.debug('Initializing git repository in "%s"', self.root)
            os.system(f"git init {self.root}")
            self.git(force=True)

        Event.PostInitCommand.fire(self)



"""Tests for coBib's config validation."""
# pylint: disable=unused-argument, redefined-outer-name

import logging
import os
from typing import Any, Generator

import pytest

from cobib.config import config
from cobib.config.config import Config

from .. import get_resource

EXAMPLE_LITERATURE = get_resource("example_literature.yaml")


def test_config_load() -> None:
    """Test loading another config file."""
    config.load(get_resource("debug.py"))
    assert config.database.file == str(EXAMPLE_LITERATURE)


def test_config_load_from_open_file() -> None:
    """Test loading another config from an open file."""
    with open(get_resource("debug.py"), "r", encoding="utf-8") as file:
        config.load(file)
    assert config.database.file == str(EXAMPLE_LITERATURE)


def test_config_load_nothing() -> None:
    """Test that nothing changes when no XDG files are present."""
    Config.XDG_CONFIG_FILE = ""
    config.load()
    # we manually call validate because load exits early
    config.validate()


def test_config_disable_load() -> None:
    """Test that loading can be disabled via environment variable."""
    prev_value = os.environ.get("COBIB_CONFIG", None)
    try:
        config.database.file = "dummy"
        os.environ["COBIB_CONFIG"] = "0"
        config.load()
        assert config.database.file == "dummy"
    finally:
        if prev_value is None:
            os.environ.pop("COBIB_CONFIG", None)
        else:
            os.environ["COBIB_CONFIG"] = prev_value


def test_config_load_from_env_var() -> None:
    """Test that loading can be configured via environment variable."""
    prev_value = os.environ.get("COBIB_CONFIG", None)
    try:
        os.environ["COBIB_CONFIG"] = get_resource("debug.py")
        config.load()
        assert config.database.file == str(EXAMPLE_LITERATURE)
    finally:
        if prev_value is None:
            os.environ.pop("COBIB_CONFIG", None)
        else:
            os.environ["COBIB_CONFIG"] = prev_value


def test_config_load_xdg() -> None:
    """Test loading a config from XDG path."""
    Config.XDG_CONFIG_FILE = get_resource("debug.py")
    config.load()
    assert config.database.file == str(EXAMPLE_LITERATURE)


def test_config_example() -> None:
    """Test that the example config matches the default values."""
    config.load(get_resource("example.py", "../src/cobib/config/"))
    assert config == Config()


def test_config_validation_failure(caplog: pytest.LogCaptureFixture) -> None:
    """Test for a `SystemExit` upon config validation failure.

    Args:
        caplog: the built-in pytest fixture.
    """
    with pytest.raises(SystemExit):
        config.load(get_resource("broken_config.py", "config"))
    assert (
        "cobib.config.config",
        logging.ERROR,
        "config.database.file should be a string.",
    ) in caplog.record_tuples


@pytest.fixture
def setup() -> Generator[Any, None, None]:
    """Setup debugging configuration.

    Yields:
        Access to the local fixture variables.
    """
    config.load(get_resource("debug.py"))
    yield setup
    config.defaults()


def test_config_validation(setup: Any) -> None:
    """Test that the initial configuration passes all validation checks.

    Args:
        setup: a local pytest fixture.
    """
    config.validate()



r"""coBib's Modify command.

This command allows you to perform bulk modification to multiple entries.
Thus, it provides faster means to apply simple edits to many entries at once, without having to open
each entry for editing one-by-one or having to edit the database file manually.

It takes a modification in the form `<field>:<value>` and will overwrite the `field` of all matching
entries with the new `value`. A simple example is the following:
```
cobib modify tags:private --selection -- Label1 Label2 ...
```
which will set the tags of all listed entries to `private`.

You can use the `--add` option to not overwrite but append to existing values of a field. `str`
fields will be concatenated with*out* any spaces, lists will be appended, and numeric fields will be
added. Any other kind of field will be converted to a `str`.

As with other commands, you can also use filters (see also `cobib.commands.list_`) rather than a
manual selection to specify the entries which to modify:
```
cobib modify tags:first_author -- ++author Rossmannek
```

As of v3.2.0 the `value` provided as part of the modification is interpreted as an "f"-string.[^1]
This means you can even use placeholder variables and perform simple operations on them. The
available variables depend on the entry which you are modifying as they are inferred from its stored
data.
Below are some examples:
```
# Rewrite the 'pages' field with a single-dash separator
cobib modify "pages:{pages.replace('--', '-')}" -- ...

# Rename an entry according to the first author's surname and year
cobib modify "label:{author.split()[1]}{year}" -- ...
```

In case you are applying a modification to your entry labels, the value of the
`cobib.config.config.ModifyCommandConfig.preserve_files` setting (added in v4.1.0) determines
whether all of your associated files will be renamed accordingly. This defaults to `False`, meaning
that they *will* be renamed. You can overwrite the value of this setting at runtime with the
`--preserve-files` and `--no-preserve-files` arguments, respectively.
I.e. the following will **not** rename your files:
```
# Rename an entry according to the first author's surname and year, but preserve the original file
cobib modify "label:{author.split()[1]}{year}" --preserve-files -- ...
```
While this command will always rename them:
```
# Rename an entry according to the first author's surname and year, and rename the original file
cobib modify "label:{author.split()[1]}{year}" --no-preserve-files -- ...
```

In combination with the regex-support for filters added during the same release, you can even unify
your database's label convention:
```
cobib modify "label:{label.replace('_', '')}" -- ++label "\D+_\d+"
```

If you happen to use an undefined variable as part of your modification, coBib will handle this
gracefully by falling back to an empty string and raising a warning:
```
cobib modify "note:{undefined}" -- ...
> [WARNING] You tried use an undefined variable. Falling back to an empty string.
> [ERROR] name 'undefined' is not defined
```

Since v3.3.0 this command also provides a "dry" mode which previews the modifications without
actually applying them.
```
cobib modify --dry <modification> -- ...
```
This is useful if you want to test large bulk modifications before running them in order to prevent
mistakes.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `m` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:modify <arguments go here>
```

[^1]: <https://docs.python.org/3/reference/lexical_analysis.html#formatted-string-literals>
"""

from __future__ import annotations

import ast
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.utils.logging import get_stream_handler
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)


class ModifyCommand(Command):
    """The Modify Command.

    This command can parse the following arguments:

        * `modification`: a string conforming to `<field>:<value>` indicating the modification that
          should be applied to all matching entries. By default, the modification will overwrite any
          existing data in the specified `field` with the new `value`. For more information about
          formatting options of `<value>` refer to the module documentation or the man-page.
        * `--dry`: run in "dry"-mode which lists modifications without applying them.
        * `-a`, `--add`: when specified, the modification's value will be added to the entry's field
          rather than overwrite it. If the field in question is numeric, the numbers will be added.
        * `--preserve-files`: skips the renaming of any associated files in case the applied
          modification acted on the entry labels. This overwrites the
          `cobib.config.config.ModifyCommandConfig.preserve_files` setting.
        * `--no-preserve-files`: does NOT skip the renaming of any associated files in case the
          applied modification acted on the entry labels. This overwrites the
          `cobib.config.config.ModifyCommandConfig.preserve_files` setting.
        * `-s`, `--selection`: when specified, the positional arguments will *not* be interpreted as
          filters but rather as a direct list of entry labels. This can be used on the command-line
          but is mainly meant for the TUIs visual selection interface (hence the name).
        * in addition to the above, you can add `filters` to specify a subset of your database for
          exporting. For more information refer to `cobib.commands.list_`.
    """

    name = "modify"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.modified_entries: List[Entry] = []
        """A list of `cobib.database.Entry` objects which were modified by this command."""

    @staticmethod
    def field_value_pair(string: str) -> Tuple[str, str]:
        """Utility method to assert the field-value pair argument type.

        This method is given to the `argparse.ArgumentParser` instance as its `type` specifier.
        An input argument is considered valid if it passes through this function without raising any
        errors, which means it conforms to the `<field>:<value>` notation.

        Args:
            string: the argument string to check.

        Returns:
            The pair of strings: `field` and `value`.
        """
        # try splitting the string into field and value, any errors will be handled by argparse
        field, *value = string.split(":")
        # NOTE: we split only the first field off in case the value contains f-string format
        # specifications
        return (field, ":".join(value))

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="modify", description="Modify subcommand parser.")
        parser.add_argument(
            "modification",
            type=ModifyCommand.field_value_pair,
            help="Modification to apply to the specified entries."
            "\nThis argument must be a string formatted as <field>:<value> where field can be any "
            "field of the entries and value can be any string which should be placed in that "
            "field. Be sure to escape this field-value pair properly, especially if the value "
            "contains spaces.",
        )
        parser.add_argument(
            "--dry",
            action="store_true",
            help="Run in 'dry'-mode, listing modifications without actually applying them.",
        )
        parser.add_argument(
            "-a",
            "--add",
            action="store_true",
            help="Adds to the modified field rather than overwriting it.",
        )
        preserve_files_group = parser.add_mutually_exclusive_group()
        preserve_files_group.add_argument(
            "--preserve-files",
            action="store_true",
            default=None,
            help="do NOT rename associated files",
        )
        preserve_files_group.add_argument(
            "--no-preserve-files",
            dest="preserve_files",
            action="store_false",
            default=None,
            help="rename associated files",
        )
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="+",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        cls.argparser = parser

    # pylint: disable=too-many-branches
    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Modify command.")

        Event.PreModifyCommand.fire(self)

        info_handler: logging.Handler
        if self.largs.dry:
            info_handler = get_stream_handler(logging.INFO)

            class ModifyInfoFilter(logging.Filter):
                """A logging filter to only print ModifyCommand INFO messages."""

                def filter(self, record: logging.LogRecord) -> bool:
                    return record.name == "cobib.commands.modify" and record.levelname == "INFO"

            info_handler.addFilter(ModifyInfoFilter())
            LOGGER.addHandler(info_handler)

        if self.largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = self.largs.filter
        else:
            LOGGER.debug("Gathering filtered list of entries to be modified.")
            filtered_entries, _ = ListCommand(*self.largs.filter).filter_entries()
            labels = [entry.label for entry in filtered_entries]

        field, value = self.largs.modification

        preserve_files = config.commands.modify.preserve_files
        if self.largs.preserve_files is not None:
            preserve_files = self.largs.preserve_files
        LOGGER.info("Associated files will%s be preserved.", "" if preserve_files else " not")

        bib = Database()

        for label in labels:  # pylint: disable=too-many-nested-blocks
            try:
                entry = bib[label]
                local_value = evaluate_as_f_string(value, {"label": label, **entry.data.copy()})

                if hasattr(entry, field):
                    prev_value = getattr(entry, field, None)
                else:
                    prev_value = entry.data.get(field, None)

                if not self.largs.add:
                    new_value = local_value
                    if local_value.isnumeric():
                        new_value = int(local_value)  # type: ignore
                else:
                    try:
                        if prev_value is None:
                            new_value = local_value
                        elif isinstance(prev_value, str):
                            new_value = prev_value + local_value
                        elif isinstance(prev_value, list):
                            new_value = prev_value + [local_value]  # type: ignore
                        elif isinstance(prev_value, int):
                            if local_value.isnumeric():
                                new_value = prev_value + int(local_value)  # type: ignore
                            else:
                                raise TypeError
                        else:
                            raise TypeError
                    except TypeError:
                        LOGGER.warning(
                            "Encountered an unexpected field type to add to. Converting the field "
                            "'%s' of entry '%s' to a simple string: '%s'.",
                            field,
                            label,
                            str(prev_value) + local_value,
                        )
                        new_value = str(prev_value) + local_value

                # guard against overwriting existing data if label gets changed
                if field == "label":
                    new_value = bib.disambiguate_label(new_value, entry)

                if new_value == prev_value:
                    LOGGER.info(
                        "New and previous values match. Skipping modification of entry '%s'.", label
                    )
                    continue

                if hasattr(entry, field):
                    if self.largs.dry:
                        LOGGER.info(
                            "%s: changing field '%s' from %s to %s",
                            entry.label,
                            field,
                            getattr(entry, field),
                            new_value,
                        )
                    setattr(entry, field, new_value)
                else:
                    if self.largs.dry:
                        LOGGER.info(
                            "%s: adding field '%s' = %s",
                            entry.label,
                            field,
                            new_value,
                        )
                    entry.data[field] = new_value

                bib.update({entry.label: entry})

                if entry.label != label:
                    bib.rename(label, entry.label)
                    if not preserve_files:
                        new_files = []
                        for file in entry.file:
                            path = RelPath(file)
                            if path.path.stem == label:
                                LOGGER.info("Also renaming associated file '%s'.", str(path))
                                target = RelPath(path.path.parent / f"{entry.label}.pdf")
                                if target.path.exists():
                                    LOGGER.warning(
                                        "Found conflicting file, not renaming '%s'.", str(path)
                                    )
                                else:
                                    if self.largs.dry:
                                        LOGGER.info(
                                            "%s: renaming associated file '%s' to '%s'",
                                            entry.label,
                                            path.path,
                                            target.path,
                                        )
                                    else:
                                        path.path.rename(target.path)
                                        new_files.append(str(target))
                                    continue
                            if not self.largs.dry:
                                new_files.append(file)
                        if not self.largs.dry and new_files:
                            entry.file = new_files

                if not self.largs.dry:
                    self.modified_entries.append(entry)
                    msg = f"'{label}' was modified."
                    LOGGER.info(msg)
            except KeyError:
                msg = f"No entry with the label '{label}' could be found."
                LOGGER.warning(msg)

        Event.PostModifyCommand.fire(self)

        if self.largs.dry:
            LOGGER.removeHandler(info_handler)
            # read also functions as a restoring method
            bib.read()
        else:
            bib.save()
            self.git()


def evaluate_ast_node(node: ast.expr, locals_: Optional[Dict[str, Any]] = None) -> str:
    """Evaluates an AST node representing an f-string.

    Args:
        node: the AST expression extracted from an f-string.
        locals_: the dictionary of local variables to be used as context for the expression
            evaluation.

    Returns:
        The evaluated AST node expression.
    """
    try:
        # pylint: disable=eval-used
        return eval(  # type: ignore
            compile(ast.Expression(node), filename="<string>", mode="eval"), locals_
        )
    except NameError as err:
        LOGGER.warning("You tried use an undefined variable. Falling back to an empty string.")
        LOGGER.error(err)
        return ""


def evaluate_as_f_string(value: str, locals_: Optional[Dict[str, Any]] = None) -> str:
    """Evaluates a string as if it were a literal f-string.

    Args:
        value: the string to be evaluated.
        locals_: the dictionary of local variables to be used as context for the expression
            evaluation.

    Returns:
        The evaluated f-string.

    Raises:
        ValueError: if an unexpected AST component type is encountered.

    References:
        <https://stackoverflow.com/a/61190684>
    """
    result: List[str] = []
    for part in ast.parse(f"f'''{value}'''").body[0].value.values:  # type: ignore
        typ = type(part)

        if typ is ast.Constant:
            result.append(part.value)

        elif typ is ast.FormattedValue:
            value = evaluate_ast_node(part.value, locals_)

            if part.conversion >= 0:
                conversions: Dict[str, Callable[[Any], str]] = {"a": ascii, "r": repr, "s": str}
                value = conversions[chr(part.conversion)](value)

            if part.format_spec:
                value = format(value, evaluate_ast_node(part.format_spec))

            result.append(str(value))

        else:
            LOGGER.warning("Unexpected AST node expression type '%s' for an f-string.", typ)
            raise ValueError

    return "".join(result)



"""coBib's Search command.

This command allows you to search your database for one or more regex-interpreted queries.
While doing so, it uses the `cobib.config.config.SearchCommandConfig.grep` tool to search associated
files, too.

As a simple example, you can query for a simple author name like so:
```
cobib search Einstein
```
You can also control whether the search is performed case *in*sensitive.
This is done via the `cobib.config.config.SearchCommandConfig.ignore_case` setting which defaults to
being `False`.
Besides this setting, you can always overwrite its value on the command line with the
`--ignore-case` (`-i` for short) and `--no-ignore-case` (`-I` for short; since v4.1.0) options.
Providing these options takes precedence over your configuration value.
Thus, the following is *always* case *in*sensitive, irrespective of your configuration.
```
cobib search --ignore-case Einstein
```
And the following is *always* sensitive to case:
```
cobib search --no-ignore-case Einstein
```

By default, the search command will provide you with 1 line of context above and below the actual
matches. You can change this number of lines by setting the `--context` option:
```
cobib search --context 4 Einstein
```
You can also permanently change the default value via the
`cobib.config.config.SearchCommandConfig.context` setting.

Finally, you can also combine the search with coBib's filtering mechanism to narrow your search down
to a subset of your database:
```
cobib search Einstein -- ++year 2020
```
Note, that we use the auxiliary `--` argument to separate the filters from the actual arguments.
While this is not strictly necessary it helps to disambiguate the origin of the arguments.

### Associated files

The search will also be performed on any associated files of your entries.
You can configure the tool which is used to perform this search via the
`cobib.config.config.SearchCommandConfig.grep` setting (defaults to `grep`).

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `/` key.

.. note::
   For more information on the searching mechanisms see also `cobib.database.Entry.search`.
"""

from __future__ import annotations

import argparse
import logging
from typing import List, Type

from rich.console import Console, ConsoleRenderable
from rich.prompt import PromptBase, PromptType
from rich.text import Text
from rich.tree import Tree
from textual.app import App
from typing_extensions import override

from cobib import __version__
from cobib.config import Event, config
from cobib.database import Entry
from cobib.ui.components import SearchView

from .base_command import ArgumentParser, Command
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)


class SearchCommand(Command):
    """The Search Command.

    This command can parse the following arguments:

        * `query`: the required positional argument corresponds to the regex-interpreted text which
          will be searched for. You may provide multiple separate queries which will be searched for
          independently.
        * `-i`, `--ignore-case`: if specified, the search will be case-**in**sensitive. This
          overwrites the `cobib.config.config.SearchCommandConfig.ignore_case` setting.
        * `-I`, `--no-ignore-case`: if specified, the search will be case-sensitive. This
          overwrites the `cobib.config.config.SearchCommandConfig.ignore_case` setting.
        * `-c`, `--context`: you can specify the number of lines of "context" which is the number of
          lines before and after the actual match to be included in the output. This is similar to
          the `-C` option of `grep`. You can configure the default value via the
          `cobib.config.config.SearchCommandConfig.context` setting.
        * in addition to the above, you can add `filters` to narrow the search down to a subset of
          your database. For more information refer to `cobib.commands.list_`.
    """

    name = "search"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.entries: List[Entry] = []
        """A filtered list of entries searched over by this command."""

        self.matches: List[List[List[str]]] = []
        """The search matches detected by this command. This is a nested list of the following
        structure: the first list level iterates over the entries; the second list level iterates
        over the matches of any given entry; the third list level iterates the (context) lines of
        any given match."""

        self.hits: int = 0
        """The number of search hits detected by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="search", description="Search subcommand parser.")
        parser.add_argument("query", type=str, nargs="+", help="text to search for")
        ignore_case_group = parser.add_mutually_exclusive_group()
        ignore_case_group.add_argument(
            "-i",
            "--ignore-case",
            action="store_true",
            default=None,
            help="ignore case for searching",
        )
        ignore_case_group.add_argument(
            "-I",
            "--no-ignore-case",
            dest="ignore_case",
            action="store_false",
            default=None,
            help="do NOT ignore case for searching",
        )
        parser.add_argument(
            "-c",
            "--context",
            type=int,
            default=config.commands.search.context,
            help="number of context lines to provide for each match",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        cls.argparser = parser

    @override
    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        search_args = []
        filter_args = []
        found_sep = False
        for arg in args:
            if arg == "--":
                found_sep = True
                continue
            if found_sep:
                filter_args.append(arg)
            else:
                search_args.append(arg)

        largs = super()._parse_args(tuple(search_args))
        largs.filter = filter_args
        return largs

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Search command.")

        Event.PreSearchCommand.fire(self)

        self.entries, _ = ListCommand(*self.largs.filter).filter_entries()

        ignore_case = config.commands.search.ignore_case
        if self.largs.ignore_case is not None:
            ignore_case = self.largs.ignore_case
        LOGGER.debug("The search will be performed case %ssensitive", "in" if ignore_case else "")

        for entry in self.entries.copy():
            matches = entry.search(self.largs.query, self.largs.context, ignore_case)
            if not matches:
                self.entries.remove(entry)
                continue

            self.matches.append(matches)
            self.hits += len(matches)

            LOGGER.debug('Entry "%s" includes %d hits.', entry.label, len(matches))

        Event.PostSearchCommand.fire(self)

    @override
    def render_porcelain(self) -> List[str]:
        output = []
        for entry, matches in zip(self.entries, self.matches):
            title = f"{entry.label}::{len(matches)}"
            output.append(title)

            for idx, match in enumerate(matches):
                for line in match:
                    output.append(f"{idx+1}::" + line.strip())

        return output

    @override
    def render_rich(self) -> ConsoleRenderable:
        ignore_case = config.commands.search.ignore_case
        if self.largs.ignore_case is not None:
            ignore_case = self.largs.ignore_case

        tree = Tree(".", hide_root=True)
        for entry, matches in zip(self.entries, self.matches):
            subtree = tree.add(
                Text.assemble(
                    (entry.label, config.commands.search.highlights.label),
                    f" - {len(matches)} match" + ("es" if len(matches) > 1 else ""),
                )
            )

            for idx, match in enumerate(matches):
                matchtree = subtree.add(str(idx + 1))
                for line in match:
                    line_text = Text(line)
                    line_text.highlight_words(
                        self.largs.query,
                        config.commands.search.highlights.query,
                        case_sensitive=not ignore_case,
                    )
                    matchtree.add(line_text)

        return tree

    @override
    def render_textual(self) -> SearchView:
        ignore_case = config.commands.search.ignore_case
        if self.largs.ignore_case is not None:
            ignore_case = self.largs.ignore_case

        tree = SearchView(".")
        for entry, matches in zip(self.entries, self.matches):
            subtree = tree.root.add(
                Text.assemble(
                    (entry.label, config.commands.search.highlights.label),
                    f" - {len(matches)} match" + ("es" if len(matches) > 1 else ""),
                ),
                # TODO: make configurable
                expand=False,
            )

            for idx, match in enumerate(matches):
                matchtree = subtree.add(
                    str(idx + 1),
                    # TODO: make configurable
                    expand=True,
                )
                for line in match:
                    line_text = Text(line)
                    line_text.highlight_words(
                        self.largs.query,
                        config.commands.search.highlights.query,
                        case_sensitive=not ignore_case,
                    )
                    matchtree.add_leaf(line_text)

        return tree



"""Tests for coBib's ExportCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import os
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Type
from zipfile import ZipFile

import pytest
from typing_extensions import override

from cobib.commands import ExportCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import get_resource
from .command_test import CommandTest

TMPDIR = Path(tempfile.gettempdir())

if TYPE_CHECKING:
    import cobib.commands


class TestExportCommand(CommandTest):
    """Tests for coBib's ExportCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return ExportCommand

    def _assert(self, args: List[str]) -> None:
        """Common assertion utility method.

        Args:
            args: the arguments which were passed to the command.
        """
        if "-b" in args:
            self._assert_bib(args)
        if "-z" in args:
            self._assert_zip(args)

    def _assert_bib(self, args: List[str]) -> None:
        """Assertion utility method for bibtex output.

        Args:
            args: the arguments which were passed to the command.
        """
        try:
            with open(TMPDIR / "cobib_test_export.bib", "r", encoding="utf-8") as file:
                with open(
                    get_resource("example_literature.bib"), "r", encoding="utf-8"
                ) as expected:
                    # NOTE: do NOT use zip_longest to omit later entries
                    for line, truth in zip(file, expected):
                        if truth[0] == "%":
                            # ignore comments
                            continue
                        assert line == truth
                    if "-s" in args:
                        with pytest.raises(StopIteration):
                            next(file)
        finally:
            # clean up file system
            os.remove(TMPDIR / "cobib_test_export.bib")

    def _assert_zip(self, args: List[str]) -> None:
        """Assertion utility method for zip output.

        Args:
            args: the arguments which were passed to the command.
        """
        try:
            with ZipFile(TMPDIR / "cobib_test_export.zip", "r") as file:
                # assert that the file does not contain a bad file
                assert file.testzip() is None
                assert file.namelist() == ["debug.py"]
                file.extract("debug.py", path=TMPDIR)
                with open(TMPDIR / "debug.py", "r", encoding="utf-8") as extracted:
                    with open(get_resource("debug.py"), "r", encoding="utf-8") as truth:
                        assert extracted.read() == truth.read()
        finally:
            try:
                # clean up file system
                os.remove(TMPDIR / "cobib_test_export.zip")
                os.remove(TMPDIR / "debug.py")
            except FileNotFoundError:
                pass

    @pytest.mark.parametrize(
        ["args"],
        [
            [["-b", str(TMPDIR / "cobib_test_export.bib")]],
            [["-b", str(TMPDIR / "cobib_test_export.bib"), "--", "++label", "einstein"]],
            [["-b", str(TMPDIR / "cobib_test_export.bib"), "-s", "--", "einstein"]],
            [["-z", str(TMPDIR / "cobib_test_export.zip")]],
            [["-z", str(TMPDIR / "cobib_test_export.zip"), "--", "++label", "einstein"]],
            [["-z", str(TMPDIR / "cobib_test_export.zip"), "-s", "--", "einstein"]],
        ],
    )
    def test_command(self, setup: Any, args: List[str]) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
        """
        if "-z" in args:
            # add a dummy file to the `einstein` entry
            entry = Database()["einstein"]
            entry.file = get_resource("debug.py")
        ExportCommand(*args).execute()
        self._assert(args)

    @pytest.mark.parametrize("dotless", [False, True])
    def test_journal_abbreviation(self, setup: Any, dotless: bool) -> None:
        """Test journal abbreviation.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            dotless: whether to abbreviate with or without punctuation.
        """
        config.utils.journal_abbreviations = [("Annalen der Physik", "Ann. Phys.")]
        args = ["-a", "-b", str(TMPDIR / "cobib_test_export.bib"), "-s", "--", "einstein"]
        if dotless:
            args.insert(1, "--dotless")
        ExportCommand(*args).execute()
        self._assert_journal_abbreviation(dotless)

    def _assert_journal_abbreviation(self, dotless: bool) -> None:
        """Assertion utility method for bibtex output.

        Args:
            dotless: whether to abbreviate with or without punctuation.
        """
        try:
            with open(TMPDIR / "cobib_test_export.bib", "r", encoding="utf-8") as file:
                for line in file:
                    if "journal" not in line:
                        continue
                    expected = "Ann Phys" if dotless else "Ann. Phys."
                    assert expected in line
        finally:
            # clean up file system
            os.remove(TMPDIR / "cobib_test_export.bib")

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        args = ["-b", str(TMPDIR / "cobib_test_export.bib"), "-s", "--", "dummy"]
        ExportCommand(*args).execute()
        assert (
            "cobib.commands.export",
            30,
            "No entry with the label 'dummy' could be found.",
        ) in caplog.record_tuples

    def test_warning_missing_output(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing output format.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        args = ["-s", "--", "einstein"]
        ExportCommand(*args).execute()
        assert ("cobib.commands.export", 40, "No output file specified!") in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["args"],
        [
            [["-b", str(TMPDIR / "cobib_test_export.bib")]],
        ],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, args: List[str]
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            args: additional arguments to pass to the command.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "export"] + args)
        self._assert(args)

    def test_event_pre_export_command(self, setup: Any) -> None:
        """Tests the PreExportCommand event."""
        args = ["-b", str(TMPDIR / "cobib_test_export.bib")]

        @Event.PreExportCommand.subscribe
        def hook(command: ExportCommand) -> None:
            command.largs.selection = True
            command.largs.filter = ["einstein"]

        assert Event.PreExportCommand.validate()

        ExportCommand(*args).execute()

        self._assert_bib(["-s"] + args + ["--", "einstein"])

    def test_event_post_export_command(self, setup: Any) -> None:
        """Tests the PostExportCommand event."""
        args = ["-b", str(TMPDIR / "cobib_test_export.bib")]

        @Event.PostExportCommand.subscribe
        def hook(command: ExportCommand) -> None:
            os.remove(str(TMPDIR / "cobib_test_export.bib"))

        assert Event.PostExportCommand.validate()

        ExportCommand(*args).execute()

        assert not os.path.exists(str(TMPDIR / "cobib_test_export.bib"))



"""Tests for coBib's path utility."""

from pathlib import Path

from cobib.utils.rel_path import RelPath


def test_relative_path() -> None:
    """Test a path which is relative to the user's home directory."""
    path = Path.home() / "dummy.txt"
    rel_path = RelPath(path)

    # the string is relative
    assert str(rel_path) == "~/dummy.txt"
    # the path property is fully-resolved
    assert rel_path.path == path
    # getting a Path-property resolves the path first
    assert rel_path.parent == Path.home()


def test_absolute_path() -> None:
    """Test a path which is not relative to the user's home directory and therefore absolute."""
    path = Path("/tmp/dummy.txt")
    abs_path = RelPath(path)

    # the string is relative
    assert str(abs_path) == "/tmp/dummy.txt"
    # the path property is fully-resolved
    assert abs_path.path == path
    # getting a Path-property resolves the path first
    assert abs_path.parent == Path("/tmp")



"""coBib's Export command.

You can use this command to export your database.
As of now only two output formats are available:
* BibLaTex files
* Zip archives

The use case of the former is obvious and can be achieved by the following:
```
cobib export --bibtex my_database.bib
```
The latter case will collect all associated files of your database in a single Zip archive:
```
cobib export --zip my_references.zip
```
This is an important feature because coBib (by design) allows you to spread associated files across
your entire file system.
With this command you can gather them in a neat package for sharing or transferring.

You can also limit the export to a subset of your database in one of two ways:

1. through filters:
   ```
   cobib export --bibtex my_private_database.bib -- ++tags private
   ```

2. through a custom selection (using `--selection` or the short-hand option `-s`)
   ```
   cobib export --selection --bibtex some_other_database.bib -- Label1 Label2
   ```

.. note::
   While this latter case is usable via the command-line interface it is more a side-effect of the
   TUI integration which provides a visual selection (defaults to the `v` key).
   The proper and arguably more useful case is the first case using filters.

Since v3.2.0, coBib supports automatic Journal abbreviations. After configuring them as explained in
`cobib.config.config.UtilsConfig.journal_abbreviations` you can leverage them during exporting like
so:
```
cobib export --abbreviate --bibtex my_database.bib
# or
cobib export --abbreviate --dotless --bibtex my_database.bib
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `x` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:export <arguments go here>
```

.. note::
   If you have already selected one or more entries, the `--selection` argument will automatically
   be added.
"""

from __future__ import annotations

import argparse
import logging
from typing import List, Type
from zipfile import ZipFile

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event
from cobib.database import Database, Entry
from cobib.parsers.bibtex import BibtexParser
from cobib.utils.journal_abbreviations import JournalAbbreviations
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)


class ExportCommand(Command):
    """The Export Command.

    This command can parse the following arguments:

        * `-b`, `--bibtex`: specifies a BibLaTex filename into which to export.
        * `-z`, `--zip`: specifies a Zip-filename into which to export associated files.
        * `-a`, `--abbreviate`: abbreviate the Journal names before exporting. See also
          `cobib.config.config.UtilsConfig.journal_abbreviations`.
        * `--dotless`: remove punctuation from the Journal abbreviations.
        * `-s`, `--selection`: when specified, the positional arguments will *not* be
          interpreted as filters but rather as a direct list of entry labels. This can
          be used on the command-line but is mainly meant for the TUIs visual selection
          interface (hence the name).
        * in addition to the above, you can add `filters` to specify a subset of your
          database for exporting. For more information refer to `cobib.commands.list_`.
    """

    name = "export"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.exported_entries: List[Entry] = []
        """A list of `cobib.database.Entry` objects which were exported by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="export", description="Export subcommand parser.")
        parser.add_argument(
            "-b", "--bibtex", type=argparse.FileType("a"), help="BibLaTeX output file"
        )
        parser.add_argument("-z", "--zip", type=argparse.FileType("a"), help="zip output file")
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        parser.add_argument(
            "-a", "--abbreviate", action="store_true", help="Abbreviate journal names"
        )
        parser.add_argument(
            "--dotless", action="store_true", help="Remove punctuation from journal abbreviations"
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Export command.")

        Event.PreExportCommand.fire(self)

        if self.largs.bibtex is None and self.largs.zip is None:
            msg = "No output file specified!"
            LOGGER.error(msg)
            return
        if self.largs.zip is not None:
            self.largs.zip = ZipFile(  # pylint: disable=consider-using-with
                self.largs.zip.name, "w"
            )

        if self.largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = self.largs.filter
            bib = Database()
            for label in labels:
                try:
                    self.exported_entries.append(bib[label])
                except KeyError:
                    msg = f"No entry with the label '{label}' could be found."
                    LOGGER.warning(msg)
        else:
            LOGGER.debug("Gathering filtered list of entries to be exported.")
            self.exported_entries, _ = ListCommand(*self.largs.filter).filter_entries()

        bibtex_parser = BibtexParser()

        for entry in self.exported_entries:
            LOGGER.info('Exporting entry "%s".', entry.label)
            if self.largs.bibtex is not None:
                if self.largs.abbreviate and "journal" in entry.data.keys():
                    entry.data["journal"] = JournalAbbreviations.abbreviate(
                        entry.data["journal"], dotless=self.largs.dotless
                    )
                entry_str = bibtex_parser.dump(entry)
                self.largs.bibtex.write(entry_str)
            if self.largs.zip is not None:
                if "file" in entry.data.keys() and entry.file is not None:
                    files = entry.file
                    if not isinstance(files, list):
                        files = [files]
                    for file in files:
                        path = RelPath(file).path
                        LOGGER.debug(
                            'Adding "%s" associated with "%s" to the zip file.', path, entry.label
                        )
                        self.largs.zip.write(path, path.name)

        Event.PostExportCommand.fire(self)

        if self.largs.bibtex is not None:
            self.largs.bibtex.close()
        if self.largs.zip is not None:
            self.largs.zip.close()



"""coBib's Parser interface."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    import cobib.database


class Parser(ABC):
    """The Parser interface.

    This interface should be implemented by all concrete parser implementations.
    If the `dump` functionality does not make sense in a specific context, an error should be logged
    but the function should return normally otherwise.
    """

    name = "base"
    """The parsers `name` is used to register itself as an input argument to the
    `cobib.commands.add.AddCommand`."""

    @abstractmethod
    def parse(self, string: str) -> Dict[str, cobib.database.Entry]:
        """Creates a new Entry from the given string.

        This method can add a URL in the special field `_download` of the
        `cobib.database.Entry.data` dictionary pointing to a file which will be downloaded and
        stored as the associated file of the parsed entry.

        Args:
            string: the input of the concrete parser type. Depending on the concrete implementation,
                this can be the actual raw data input or a path to a file containing the raw data.
                It is the responsibility of the concrete implementation to deal with all of these
                scenarios.

        Returns:
            An `OrderedDict` mapping labels to `cobib.database.Entry` instances generated from the
            raw data.
        """

    @abstractmethod
    def dump(self, entry: cobib.database.Entry) -> Optional[str]:
        """Dumps an entry in the parsers format.

        Args:
            entry: the `cobib.database.Entry` to be dumped.

        Returns:
            A `str`-representation of the entry in the concrete parsers format. If the conversion
            does not make sense for a certain concrete implementation, `None` is returned and an
            error should be logged. This function should *not* raise an actual error or exit
            prematurely.
        """



"""coBib's Delete command.

This command can be used to deleted entries from the database.
```
cobib delete <label 1> [<label 2> ...]
```

When you delete an entry, the value of `cobib.config.config.DeleteCommandConfig.preserve_files`
setting (added in v4.1.0) determines whether the associated files will be deleted, too. This
defaults to `False`, meaning that they *will* be deleted. You can overwrite the value of this
setting at runtime with the `--preserve-files` and `--no-preserve-files` arguments, respectively.
I.e. the following will **not** delete your files:
```
cobib delete --preserve-files <label 1> [<label 2> ...]
```
While this command will always delete them:
```
cobib delete --no-preserve-files <label 1> [<label 2> ...]
```

As of coBib v4.1.0, the user will be asked to confirm the deletion via an interactive prompt. This
can be disabled by setting `cobib.config.config.DeleteCommandConfig.confirm` to `False`.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `d` key.
"""

from __future__ import annotations

import logging
import os
from functools import wraps
from typing import Callable, Set, Type, cast

from rich.console import Console
from rich.prompt import Confirm, InvalidResponse, PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class DeleteCommand(Command):
    """The Delete Command.

    This command can parse the following arguments:

        * `labels`: one (or multiple) labels of the entries to be deleted.
        * `--preserve-files`: skips the deletion of any associated files. This overwrites the
          `cobib.config.config.DeleteCommandConfig.preserve_files` setting.
        * `--no-preserve-files`: does NOT skip the deletion of any associated files. This overwrites
          the `cobib.config.config.DeleteCommandConfig.preserve_files` setting.
    """

    name = "delete"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        if prompt is None:
            prompt = Confirm  # type: ignore[assignment]

        super().__init__(*args, console=console, prompt=prompt)

        self.deleted_entries: Set[str] = set()
        """A set of labels which were deleted by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="delete", description="Delete subcommand parser.")
        parser.add_argument("labels", type=str, nargs="+", help="labels of the entries")
        preserve_files_group = parser.add_mutually_exclusive_group()
        preserve_files_group.add_argument(
            "--preserve-files",
            action="store_true",
            default=None,
            help="do not delete associated files",
        )
        preserve_files_group.add_argument(
            "--no-preserve-files",
            dest="preserve_files",
            action="store_false",
            default=None,
            help="delete associated files",
        )
        cls.argparser = parser

    @override
    async def execute(self) -> None:  # type: ignore[override]
        # pylint: disable=invalid-overridden-method
        LOGGER.debug("Starting Delete command.")

        Event.PreDeleteCommand.fire(self)

        preserve_files = config.commands.delete.preserve_files
        if self.largs.preserve_files is not None:
            preserve_files = self.largs.preserve_files
        LOGGER.info("Associated files will%s be preserved.", "" if preserve_files else " not")

        # pylint: disable=line-too-long
        self.prompt.process_response = self._wrap_prompt_process_response(  # type: ignore[method-assign]
            self.prompt.process_response  # type: ignore[assignment]
        )

        bib = Database()
        for label in self.largs.labels:
            try:
                LOGGER.debug("Attempting to delete entry '%s'.", label)

                if config.commands.delete.confirm:
                    prompt_text = f"Are you sure you want to delete the entry '{label}'?"

                    if self.prompt is not Confirm:
                        res = await self.prompt.ask(  # type: ignore[call-overload]
                            prompt_text,
                            choices=["y", "n"],
                            default="y",
                            console=cast(App[None], self.console),
                        )
                    else:
                        res = self.prompt.ask(
                            prompt_text,
                            default=True,
                            console=cast(Console, self.console),
                        )

                    if not res:
                        continue

                entry = bib.pop(label)
                if not preserve_files:
                    for file in entry.file:
                        path = RelPath(file)
                        try:
                            LOGGER.debug("Attempting to remove associated file '%s'.", str(path))
                            os.remove(path.path)
                        except FileNotFoundError:
                            pass

                self.deleted_entries.add(label)
            except KeyError:
                pass

        self.prompt.process_response = (  # type: ignore[method-assign]
            self.prompt.process_response.__wrapped__  # type: ignore[attr-defined]
        )

        Event.PostDeleteCommand.fire(self)
        bib.save()

        self.git()

        for label in self.deleted_entries:
            msg = f"'{label}' was removed from the database."
            LOGGER.info(msg)

    @staticmethod
    def _wrap_prompt_process_response(
        func: Callable[[PromptBase[PromptType], str], PromptType]
    ) -> Callable[[PromptBase[PromptType], str], PromptType]:
        """A method to wrap a `PromptBase.process_response` method.

        This method wraps a `PromptBase.process_response` method in order to handle a user's request
        for additional help.

        Args:
            func: the `PromptBase.process_response` method to be wrapped.

        Returns:
            The wrapped `PromptBase.process_response` method.
        """

        @override
        @wraps(func)
        def process_response(prompt: PromptBase[PromptType], value: str) -> PromptType:
            return_value: PromptType = func(prompt, value)

            if isinstance(return_value, bool):
                return return_value  # type: ignore[return-value]

            if cast(str, return_value).strip().lower() == "y":
                return True  # type: ignore[return-value]

            if cast(str, return_value).strip().lower() == "n":
                return False  # type: ignore[return-value]

            raise InvalidResponse("You may only provide 'y' or 'n' as a valid response.")

        return process_response



"""coBib's Edit command.

This command can be used to manually edit database entries in their easily-readable YAML format.
To get started, simply type:
```
cobib edit <label>
```
which will open the YAML-formatted version of the specified `cobib.database.Entry` for editing.

You can configure which editor will be used via the `cobib.config.config.EditCommandConfig.editor`
setting which will default to using your `$EDITOR` environment setting (and fall back to `vim` if
that is not set).

You can even add entirely new entries to the database by specifying an unused entry label *and*
adding the `--add` command-line argument:
```
cobib edit --add <new label>
```
This entry will be entirely empty except for the one field which is always present:
    * `ENTRYTYPE`: set to the default value configured via
      `cobib.config.config.EditCommandConfig.default_entry_type`.

If you change the label of the entry during editing, the value of the
`cobib.config.config.EditCommandConfig.preserve_files` setting (added in v4.1.0) determines whether
the associated files will be renamed automatically. This defaults to `False`, meaning that they
*will* be renamed. You can overwrite the value of this setting at runtime with the
`--preserve-files` and `--no-preserve-files` arguments, respectively.
I.e. the following will **not** rename your files:
```
cobib edit --preserve-files <label>
```
While this command will always rename them:
```
cobib edit --no-preserve-files <label>
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `e` key.
If you want to add a new entry manually, you will have to enter the prompt (defaults to `:`) and
then type out the command mentioned above:
```
:edit --add <new label>
```
"""

from __future__ import annotations

import logging
import os
import tempfile
from pathlib import Path
from typing import Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.parsers.yaml import YAMLParser
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class EditCommand(Command):
    """The Edit Command.

    This command can parse the following arguments:

        * `label`: the label of the entry to edit.
        * `-a`, `--add`: if specified, allows adding a new entry for a non-existent label. The
          default entry type of this new entry can be configured via
          `cobib.config.config.EditCommandConfig.default_entry_type`.
        * `--preserve-files`: skips the renaming of any associated files in case you manually rename
          the entry label during editing. This overwrites the
          `cobib.config.config.EditCommandConfig.preserve_files` setting.
        * `--no-preserve-files`: does NOT skip the renaming of any associated files in case you
          manually rename the entry label during editing. This overwrites the
          `cobib.config.config.EditCommandConfig.preserve_files` setting.
    """

    name = "edit"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.new_entry: Entry
        """A `cobib.database.Entry` instance edited by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="edit", description="Edit subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")
        parser.add_argument(
            "-a",
            "--add",
            action="store_true",
            help="if specified, will add a new entry for unknown labels",
        )
        preserve_files_group = parser.add_mutually_exclusive_group()
        preserve_files_group.add_argument(
            "--preserve-files",
            action="store_true",
            default=None,
            help="do NOT rename associated files",
        )
        preserve_files_group.add_argument(
            "--no-preserve-files",
            dest="preserve_files",
            action="store_false",
            default=None,
            help="rename associated files",
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Edit command.")

        Event.PreEditCommand.fire(self)

        yml = YAMLParser()

        bib = Database()

        try:
            entry = bib[self.largs.label]
            prv = yml.dump(entry)
            if self.largs.add:
                LOGGER.warning(
                    "Entry '%s' already exists! Ignoring the `--add` argument.", self.largs.label
                )
                self.largs.add = False
        except KeyError:
            # No entry for given label found
            if self.largs.add:
                # add a new entry for the unknown label
                entry = Entry(
                    self.largs.label,
                    {"ENTRYTYPE": config.commands.edit.default_entry_type},
                )
                prv = yml.dump(entry)
            else:
                msg = (
                    f"No entry with the label '{self.largs.label}' could be found."
                    "\nUse `--add` to add a new entry with this label."
                )
                LOGGER.error(msg)
                return
        if prv is None:
            # No entry found to be edited. This should never occur unless the YAMLParser experiences
            # an unexpected error.
            return

        LOGGER.debug("Creating temporary file.")
        with tempfile.NamedTemporaryFile(mode="w+", prefix="cobib-", suffix=".yaml") as tmp_file:
            tmp_file_name = tmp_file.name
            tmp_file.write(prv)
            tmp_file.flush()
            LOGGER.debug('Starting editor "%s".', config.commands.edit.editor)
            status = os.system(config.commands.edit.editor + " " + tmp_file.name)
            assert status == 0
            LOGGER.debug("Editor finished successfully.")
            new_entries = YAMLParser().parse(tmp_file.name)
            self.new_entry = list(new_entries.values())[0]
        assert not Path(tmp_file_name).exists()
        if entry == self.new_entry and not self.largs.add:
            LOGGER.info("No changes detected.")
            return

        bib.update({self.new_entry.label: self.new_entry})

        preserve_files = config.commands.edit.preserve_files
        if self.largs.preserve_files is not None:
            preserve_files = self.largs.preserve_files
        LOGGER.info("Associated files will%s be preserved.", "" if preserve_files else " not")

        if self.new_entry.label != self.largs.label:
            bib.rename(self.largs.label, self.new_entry.label)
            if not preserve_files:
                new_files = []
                for file in self.new_entry.file:
                    path = RelPath(file)
                    if path.path.stem == self.largs.label:
                        LOGGER.info("Also renaming associated file '%s'.", str(path))
                        target = RelPath(path.path.parent / f"{self.new_entry.label}.pdf")
                        if target.path.exists():
                            LOGGER.warning("Found conflicting file, not renaming '%s'.", str(path))
                        else:
                            path.path.rename(target.path)
                            new_files.append(str(target))
                            continue
                    new_files.append(file)
                self.new_entry.file = new_files

        Event.PostEditCommand.fire(self)
        bib.save()

        self.git()

        msg = f"'{self.largs.label}' was successfully edited."
        LOGGER.info(msg)



"""coBib's database module.

coBib uses a plain-test YAML file to store the bibliographic database.
This module contains the components which represent this database at runtime.
"""

from .database import Database
from .entry import Entry

__all__ = [
    "Database",
    "Entry",
]



"""coBib's Add command.

This command allows you to add new entries to your database.
You have two options on how to use this command.

### 1. Parser-based addition

All parsers available in `cobib.parsers` will be registered (at runtime) in a *mutually exclusive*
group of keyword arguments.
This means, that you can add a new entry based on any of them by using the keyword argument
according to the parsers name.
For example:
```
cobib add --bibtex some_biblatex_file.bib
cobib add --arxiv <some arXiv ID>
cobib add --doi <some DOI>
cobib add --isbn <some ISBN>
cobib add --url <some URL>
cobib add --yaml some_cobib_style_yaml_file.yaml
```

Note, that you cannot combine multiple parsers within a single command execution (for obvious
reasons).

Furthermore, most of these parsers will set the entry label automatically.
If you would like to prevent that and specify a custom label directly, you can add the `--label`
keyword argument to your command like so:
```
cobib add --doi <some DOI> --label "MyLabel"
```

.. note::
   Since this command adds new entries to the database, its outcome can be affected by your
   `cobib.config.config.DatabaseConfig` settings. In particular, pay attention to the
   `cobib.config.config.EntryStringifyConfig` settings which affect how entries are converted
   to/from strings. In particular, the following setting will affect how multiple files are split
   into a list of files:
   ```
   config.database.stringify.list_separator.file = ", "
   ```
   The above will separate file paths using `, ` but if you use a different separator (for example
   `;`) be sure to update this setting accordingly.

### 2. Manual entry addition

If you want to add a new entry manually, you *must* omit any of the parser keyword arguments and
instead specify a new label like so:
```
cobib add --label <some new label>
```
This will trigger the `cobib.commands.edit.EditCommand` for a manual addition.
However, the benefit of using this through the `AddCommand` is the availability of the following
additional options which are always available (i.e. also in combination with the parser keyword
arguments, above).

### Additional Options

You can directly associate your new entry with one or multiple files by doing the following:
```
cobib add --doi <some DOI> --file /path/to/a/file.pdf [/path/to/another/file.dat ...]
```

You can also specify `cobib.database.Entry.tags` using *positional* arguments like so:
```
cobib add --doi <some DOI> -- tag1 "multi-word tag2" ...
```

Since v3.2.0 coBib attempts to download PDF files for newly added entries (if the corresponding
parser supports this feature). The default location where this file will be stored can be configured
via `cobib.config.config.FileDownloaderConfig.default_location`, but it can be changed at runtime
using the `--path` argument like so:
```
cobib add --path <some custom path> --arxiv <some arXiv ID>
```
Since v4.1.0 you can suppress the automatic download via the
`cobib.config.config.AddCommandConfig.skip_download` setting. It defaults to `False` meaning that
the download will be attempted.
If you want to manually overwrite the configuration setting you can do so with the `--skip-download`
and `--force-download` arguments, respectively.
I.e. the following will **not** attempt the automatic download:
```
cobib add --skip-download --arxiv <some arXiv ID>
```
While the next command will always attempt the automatic download:
```
cobib add --force-download --arxiv <some arXiv ID>
```

Since v4.0.0 coBib will ask you what to do when encountering a conflict at runtime. That means, when
a label already exists in your database, you have various choices how to handle it:

1. you can `keep` the existing entry and skip the addition of the new one.
   .. note::
      This is equivalent to the old `--skip-existing` argument (added in v3.3.0) which is now
      deprecated and will be removed in a future version.

2. you can `replace` the existing entry.

3. you can `update` the existing entry.
   ```
   cobib add --doi <some DOI> --disambiguation update --label <some existing label>```

   This will take the existing entry and combine it with all new information found in the freshly
   added entry. Existing fields will be overwritten. If you have an automatically downloaded file
   associated with this entry, that will also be overwritten.
   This feature is especially useful if you want to update an entry which you previously added from
   the arXiv with its newly published version.
   .. note::
      This is equivalent to the `--update` argument (added in v3.3.0) which is now deprecated and
      will be removed in a future version.

4. you can `disambiguate` the entries. This will be done based on the
   `cobib.config.config.DatabaseFormatConfig.label_suffix` setting. It defaults to appending `_a`,
   `_b`, etc. to the label in order to differentiate (disambiguate) it from the already existing
   one.

When running into such a case, coBib will generate a side-by-side comparison of the existing and new
entry and give you the choice of how to proceed. The default choice is to *cancel* the addition
(i.e. `keep`) in order to prevent data loss.

If you already know that you will run into this case, you can bypass the prompt via the
`--disambiguation` argument and provide the intended answer ahead of time, for example like so:
```
cobib add --doi <some DOI> --disambiguation replace --label <some existing label>
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `a` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:add <arguments go here>
```
"""

from __future__ import annotations

import argparse
import asyncio
import inspect
import logging
from collections import OrderedDict
from functools import wraps
from typing import Callable, Dict, Type, cast

from rich.console import Console
from rich.prompt import InvalidResponse, Prompt, PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib import parsers
from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.parsers import BibtexParser
from cobib.utils.diff_renderer import Differ
from cobib.utils.file_downloader import FileDownloader
from cobib.utils.journal_abbreviations import JournalAbbreviations

from .base_command import ArgumentParser, Command
from .edit import EditCommand
from .modify import evaluate_as_f_string

LOGGER = logging.getLogger(__name__)


class AddCommand(Command):
    """The Add Command.

    This command can parse the following arguments:

        * `-l`, `--label`: the label to give to the new entry.
        * `-d`, `--disambiguation`: hard-codes the reply to be used if a disambiguation prompt would
          occur.
        * `-f`, `--file`: one or multiple files to associate with this entry. This data will be
          stored in the `cobib.database.Entry.file` property.
        * `-p`, `--path`: the path to store the downloaded associated file in. This can be used to
          overwrite the `cobib.config.config.FileDownloaderConfig.default_location`.
        * `--skip-download`: skips the automatic download of an associated file.
        * `--force-download`: forces the automatic download of an associated file.
        * `--skip-existing`: **DEPRECATED** use `--disambiguation keep` instead!
        * `-u`, `--update`: **DEPRECATED** use `--disambiguation update` instead!
        * in addition to the options above, a *mutually exclusive group* of keyword arguments for
          all available `cobib.parsers` are registered at runtime. Please check the output of
          `cobib add --help` for the exact list.
        * any *positional* arguments (i.e. those, not preceded by a keyword) are interpreted as tags
          and will be stored in the `cobib.database.Entry.tags` property.
    """

    name = "add"

    _avail_parsers = {
        cls.name: cls for _, cls in inspect.getmembers(parsers) if inspect.isclass(cls)
    }
    """The available parsers."""

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.new_entries: Dict[str, Entry] = OrderedDict()
        """An `OrderedDict` mapping labels to `cobib.database.Entry` instances which were added by
        this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="add", description="Add subcommand parser.")
        parser.add_argument("-l", "--label", type=str, help="the label for the new database entry")
        parser.add_argument(
            "-f",
            "--file",
            type=str,
            nargs="+",
            action="extend",
            help="files associated with this entry",
        )
        parser.add_argument("-p", "--path", type=str, help="the path for the downloaded file")
        parser.add_argument(
            "--disambiguation",
            type=str,
            choices=["keep", "replace", "update", "disambiguate"],
            help="the reply in case of a disambiguation prompt",
        )
        skip_download_group = parser.add_mutually_exclusive_group()
        skip_download_group.add_argument(
            "--skip-download",
            action="store_true",
            default=None,
            help="skip the automatic download of an associated file",
        )
        skip_download_group.add_argument(
            "--force-download",
            dest="skip_download",
            action="store_false",
            default=None,
            help="force the automatic download of an associated file",
        )
        parser.add_argument(
            "--skip-existing",
            action="store_true",
            help="DEPRECATED: use '--disambiguation keep' instead!",
        )
        parser.add_argument(
            "-u",
            "--update",
            action="store_true",
            help="DEPRECATED: use '--disambiguation update' instead!",
        )
        group_add = parser.add_mutually_exclusive_group()
        for name in cls._avail_parsers.keys():
            try:
                group_add.add_argument(
                    f"-{name[0]}", f"--{name}", type=str, help=f"{name} object identfier"
                )
            except argparse.ArgumentError:
                try:
                    group_add.add_argument(f"--{name}", type=str, help=f"{name} object identfier")
                except argparse.ArgumentError:
                    continue
        parser.add_argument(
            "tags",
            nargs=argparse.REMAINDER,
            help="A list of space-separated tags to associate with this entry."
            "\nYou can use quotes to specify tags with spaces in them.",
        )

        cls.argparser = parser

    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        largs = super()._parse_args(args)

        if largs.skip_existing:
            msg = (
                "The '--skip-existing' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation keep'."
            )
            LOGGER.warning(msg)
            largs.disambiguation = "keep"

        if largs.update:
            msg = (
                "The '--update' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation update'."
            )
            LOGGER.warning(msg)
            largs.disambiguation = "update"

        return largs

    # pylint: disable=invalid-overridden-method,too-many-statements,too-many-branches,too-many-locals
    @override
    async def execute(self) -> None:  # type: ignore[override]
        LOGGER.debug("Starting Add command.")

        Event.PreAddCommand.fire(self)

        edit_entries = False
        for name, cls in AddCommand._avail_parsers.items():
            string = getattr(self.largs, name, None)
            if string is None:
                continue
            LOGGER.debug("Adding entries from %s: '%s'.", name, string)
            self.new_entries = cls().parse(string)
            break
        else:
            if self.largs.label is not None:
                LOGGER.warning(
                    "No input to parse. Creating new entry '%s' manually.", self.largs.label
                )
                self.new_entries = {
                    self.largs.label: Entry(
                        self.largs.label,
                        {"ENTRYTYPE": config.commands.edit.default_entry_type},
                    )
                }
                edit_entries = True
            else:
                msg = "Neither an input to parse nor a label for manual creation specified!"
                LOGGER.error(msg)
                return

        if self.largs.label is not None:
            assert len(self.new_entries.values()) == 1
            for value in self.new_entries.copy().values():
                # logging done by cobib/database/entry.py
                value.label = self.largs.label
            self.new_entries = OrderedDict(
                (self.largs.label, value) for value in self.new_entries.values()
            )
        else:
            formatted_entries = OrderedDict()
            for label, value in self.new_entries.items():
                formatted_label = evaluate_as_f_string(
                    config.database.format.label_default, {"label": label, **value.data.copy()}
                )
                value.label = formatted_label
                formatted_entries[formatted_label] = value
            self.new_entries = formatted_entries

        if self.largs.file is not None:
            assert len(self.new_entries.values()) == 1
            for value in self.new_entries.values():
                # logging done by cobib/database/entry.py
                value.file = self.largs.file

        if self.largs.tags != []:
            assert len(self.new_entries.values()) == 1
            for value in self.new_entries.values():
                # logging done by cobib/database/entry.py
                value.tags = self.largs.tags

        skip_download = config.commands.add.skip_download
        if self.largs.skip_download is not None:
            skip_download = self.largs.skip_download
        LOGGER.info("Automatic file download will%s be attempted.", " not" if skip_download else "")

        bib = Database()
        existing_labels = set(bib.keys())

        for lbl, entry in self.new_entries.copy().items():
            overwrite_file = False
            # check if label already exists
            if lbl in existing_labels:
                # if it does, we have multiple cases to differentiate:
                if edit_entries:
                    # the user tried to manually add an entry which already exists, point them to
                    # the edit command instead
                    msg = (
                        f"You tried to add the '{lbl}' entry manually, but it already exists, "
                        f"please use `cobib edit {lbl}` instead!"
                    )
                    LOGGER.warning(msg)
                    continue
                # in all other cases, we enter an interactive prompt to let the user decide how
                # to proceed
                msg = f"You tried to add a new entry '{lbl}' which already exists!"
                LOGGER.warning(msg)
                res = self.largs.disambiguation
                # if the user provided the reply at runtime using the --disambiguation argument, the
                # following condition is not met
                if res is None:
                    parser = BibtexParser()
                    left = parser.dump(bib[lbl])
                    right = parser.dump(entry)
                    diff = Differ(left, right)
                    diff.compute()
                    table = diff.render("bibtex")

                    prompt_text = "How would you like to handle this conflict?"
                    choices = ["keep", "replace", "update", "disambiguate", "help"]
                    default = "keep"

                    # pylint: disable=line-too-long
                    self.prompt.process_response = self._wrap_prompt_process_response(  # type: ignore[method-assign]
                        self.prompt.process_response  # type: ignore[assignment]
                    )

                    if self.prompt is not Prompt:

                        @wraps(self.prompt.pre_prompt)
                        def pre_prompt(_prompt: PromptBase[PromptType]) -> None:
                            _prompt.console.push_screen("input")  # type: ignore[attr-defined]
                            _prompt.console.print(table)  # pylint: disable=cell-var-from-loop

                        # pylint: disable=line-too-long
                        self.prompt.pre_prompt = pre_prompt  # type: ignore[assignment,method-assign]

                        res = await self.prompt.ask(  # type: ignore[call-overload]
                            prompt_text,
                            choices=choices,
                            default=default,
                            console=cast(App[None], self.console),
                        )

                        self.prompt.pre_prompt = (  # type: ignore[method-assign]
                            self.prompt.pre_prompt.__wrapped__  # type: ignore[attr-defined]
                        )
                    else:
                        self.console.print(table)  # type: ignore[union-attr]

                        res = self.prompt.ask(
                            prompt_text,
                            choices=choices,
                            default=default,
                            console=cast(Console, self.console),
                        )

                    self.prompt.process_response = (  # type: ignore[method-assign]
                        self.prompt.process_response.__wrapped__  # type: ignore[attr-defined]
                    )

                if res == "update":
                    msg = f"Updating the already existing entry '{lbl}' with the new data."
                    LOGGER.info(msg)
                    existing_data = bib[lbl].data.copy()
                    existing_data.update(entry.data)
                    entry.data = existing_data.copy()
                    overwrite_file = True
                elif res == "disambiguate":
                    msg = (
                        "The label will be disambiguated based on the configuration option: "
                        "config.database.format.label_suffix"
                    )
                    LOGGER.warning(msg)
                    new_label = bib.disambiguate_label(lbl, entry)
                    entry.label = new_label
                    self.new_entries[new_label] = entry
                    self.new_entries.pop(lbl)
                elif res == "replace":
                    msg = f"Overwriting the already existing entry '{lbl}' with the new data."
                    LOGGER.info(msg)
                elif res == "keep":
                    msg = f"Skipping addition of the already existing entry '{lbl}'."
                    LOGGER.info(msg)
                    self.new_entries.pop(lbl)
                    continue

            # download associated file (if requested)
            if "_download" in entry.data.keys():
                if skip_download:
                    entry.data.pop("_download")
                else:
                    task = asyncio.create_task(
                        FileDownloader().download(
                            entry.data.pop("_download"),
                            entry.label,
                            folder=self.largs.path,
                            overwrite=overwrite_file,
                        )
                    )
                    path = await task
                    if path is not None:
                        entry.file = str(path)  # type: ignore[assignment]
            # check journal abbreviation
            if "journal" in entry.data.keys():
                entry.data["journal"] = JournalAbbreviations.elongate(entry.data["journal"])

        Event.PostAddCommand.fire(self)

        bib.update(self.new_entries)
        if edit_entries:
            EditCommand(self.largs.label).execute()

        bib.save()

        self.git()

        for label in self.new_entries:
            msg = f"'{label}' was added to the database."
            LOGGER.info(msg)

    @staticmethod
    def _wrap_prompt_process_response(
        func: Callable[[PromptBase[PromptType], str], PromptType]
    ) -> Callable[[PromptBase[PromptType], str], PromptType]:
        """A method to wrap a `PromptBase.process_response` method.

        This method wraps a `PromptBase.process_response` method in order to handle a user's request
        for additional help.

        Args:
            func: the `PromptBase.process_response` method to be wrapped.

        Returns:
            The wrapped `PromptBase.process_response` method.
        """

        @override
        @wraps(func)
        def process_response(prompt: PromptBase[PromptType], value: str) -> PromptType:
            return_value: PromptType = func(prompt, value)

            if return_value == "help":
                LOGGER.debug("User requested help.")
                raise InvalidResponse(
                    "[yellow]A conflict between an existing (left) and newly added entry (right) "
                    "occurred. These are your options:\n"
                    "  'keep' the existing entry and discard the new addition (default)\n"
                    "  'replace' the existing entry with the new one\n"
                    "  'update' the existing entry with the new data\n"
                    "  'disambiguate' the new entry from the existing one by adding a label suffix"
                )

            return return_value

        return process_response



"""coBib's shell helpers.

This module provides a variety of shell helper utilities.
"""
# pylint: disable=unused-argument

from __future__ import annotations

import argparse
import contextlib
import inspect
import logging
from io import StringIO
from typing import List, Set, Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App

from .rel_path import RelPath

LOGGER = logging.getLogger(__name__)


def list_commands(*args: str) -> List[str]:
    """Lists all available subcommands.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of available commands.
    """
    msg = "The _list_commands shell helper utility is deprecated and will be removed in the future!"
    LOGGER.warning(msg)

    # pylint: disable=import-outside-toplevel
    from cobib import commands

    return [cls.name for _, cls in inspect.getmembers(commands) if inspect.isclass(cls)]


def list_labels(*args: str) -> List[str]:
    """List all available labels in the database.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of all labels.
    """
    msg = "The _list_labels shell helper utility is deprecated and will be removed in the future!"
    LOGGER.warning(msg)

    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    labels = list(Database().keys())
    return labels


def list_filters(*args: str) -> Set[str]:
    """Lists all field names available for filtering.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of all available filters.
    """
    msg = "The _list_filters shell helper utility is deprecated and will be removed in the future!"
    LOGGER.warning(msg)

    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    filters: Set[str] = {"label"}
    for entry in Database().values():
        filters.update(entry.data.keys())
    return filters


def example_config(*args: str) -> List[str]:
    """Shows the (well-commented) example configuration.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The lines of the example config file.
    """
    root = RelPath(__file__).parent.parent
    with open(root / "config/example.py", "r", encoding="utf-8") as file:
        return [line.strip() for line in file.readlines()]


class LintFormatter(logging.Formatter):
    """A custom logging.Formatter."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        # noqa: D107
        super().__init__(*args, **kwargs)

        self.dirty_entries: Set[str] = set()

        # pylint: disable=import-outside-toplevel
        from cobib.config import config

        self._database_path = RelPath(config.database.file)

        with open(self._database_path.path, "r", encoding="utf-8") as database:
            self._raw_database = database.readlines()

    def format(self, record: logging.LogRecord) -> str:
        """Format's the LogRecord.

        This custom Formatter uses the LogRecord's attributes to determine from which line of the
        raw database a formatting information was raised. The corresponding line number is used in
        conjunction with the actual message of the LogRecord for the formatting.

        Args:
            record: the LogRecord to be formatted.

        Returns:
            A string encoding the LogRecord's information.
        """
        try:
            entry = record.entry  # type: ignore[attr-defined]
            field = record.field  # type: ignore[attr-defined]
            self.dirty_entries.add(entry)
            raw_db = enumerate(self._raw_database)
            _, line = next(raw_db)
            while not line.startswith(entry):
                _, line = next(raw_db)
            while not line.strip().startswith(field):
                line_no, line = next(raw_db)
            return f"{self._database_path}:{line_no+1} {record.getMessage()}"
        except AttributeError:
            return ""


def lint_database(*args: str) -> List[str]:
    """Lints the users database.

    Args:
        args: a sequence of additional arguments used for the execution. Currently, only a single
            optional value is allowed:
                * `-f`, `--format`: if specified, the database will be formatted to automatically
                    resolve all lint messages.

    Returns:
        The list of INFO log messages raised upon Entry initialization.
    """
    parser = argparse.ArgumentParser(prog="lint_database", description="A database format linter.")
    parser.add_argument(
        "-f",
        "--format",
        action="store_true",
        help="Automatically format database to conform with linter.",
    )
    largs = parser.parse_args(args)

    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    output = StringIO()

    handler = logging.StreamHandler(output)
    handler.setLevel(logging.INFO)
    handler.addFilter(logging.Filter("cobib.database.entry"))

    formatter = LintFormatter()
    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    if root_logger.getEffectiveLevel() > logging.INFO:
        # overwriting all existing handlers with this local one
        root_logger.handlers = [handler]
        root_logger.setLevel(logging.INFO)
    else:
        # appending new handler
        root_logger.addHandler(handler)

    # trigger database reading to cause lint messages upon entry-construction
    Database.read()

    lint_messages = output.getvalue().split("\n")

    root_logger.removeHandler(handler)

    if all(not msg for msg in lint_messages):
        return ["Congratulations! Your database triggers no lint messages."]

    if largs.format:
        for label in formatter.dirty_entries:
            # we exploit the rename method to register all dirty entries for re-writing
            Database().rename(label, label)

        Database.save()

        # pylint: disable=import-outside-toplevel
        from cobib.commands.base_command import Command

        # generate automatic git commit
        class LintCommand(Command):
            """The linting command."""

            name = "lint"

            def __init__(
                self,
                *args: str,
                console: Console | App[None] | None = None,
                prompt: Type[PromptBase[PromptType]] | None = None,
            ) -> None:
                # pylint: disable=super-init-not-called
                self.largs = largs

            @classmethod
            def init_argparser(cls) -> None:
                pass

            def execute(self):  # type: ignore
                pass

        LintCommand().git()

        return ["The following lint messages have successfully been resolved:"] + lint_messages

    return lint_messages


def unify_labels(*args: str) -> List[str]:
    """Unifies all labels of the database according to `config.database.format.label_default`.

    Without the `--apply` argument this will only print the modification which would be performed!

    Args:
        args: a sequence of additional arguments used for the execution. Currently, only a single
            optional value is allowed:
                * `-a`, `--apply`: if specified, the label unification will actually be applied. The
                    default is to run in "dry"-mode which only prints the modifications.

    Returns:
        The list of INFO log messages raised upon label unification.
    """
    parser = argparse.ArgumentParser(prog="unify_labels", description="Label unification")
    parser.add_argument(
        "-a",
        "--apply",
        action="store_true",
        help="Actually apply the modifications rather than run in 'dry'-mode",
    )
    largs = parser.parse_args(args)

    # pylint: disable=import-outside-toplevel
    from cobib.commands import ModifyCommand
    from cobib.config import config

    modify_args = [
        "--dry",
        f"label:{config.database.format.label_default}",
        "--",
        "--label",
        "some_non_existend_label",  # this ensures that the command gets run on the entire database
    ]
    if largs.apply:
        modify_args = modify_args[1:]

    with contextlib.redirect_stderr(StringIO()) as out:
        cmd = ModifyCommand(*modify_args)
        cmd.execute()

    return out.getvalue().strip().split("\n")



"""coBib's `logging.Handler` for interactive UIs.

This handler is used by the `cobib.ui.tui.TUI` to render logging messages as
`cobib.ui.components.popup.Popup` instances.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from typing_extensions import override

from .popup import Popup

if TYPE_CHECKING:
    from ..tui import TUI


class PopupLoggingHandler(logging.Handler):
    """coBib's `logging.Handler` for interactive UIs."""

    def __init__(self, app: TUI, level: int = logging.INFO) -> None:
        """Initializes the handler.

        Args:
            app: the running TUI instance.
            level: the default logging level to be displayed.
        """
        super().__init__(level=level)

        self._app = app

        formatter = logging.Formatter(fmt="[%(levelname)s] %(message)s")
        self.setFormatter(formatter)

        for handler in self._app.root_logger.handlers[:]:
            if isinstance(handler, logging.StreamHandler):
                self._app.root_logger.removeHandler(handler)
                handler.close()

        self._app.root_logger.addHandler(self)

    @override
    def emit(self, record: logging.LogRecord) -> None:
        self._app.print(Popup(self.format(record), level=record.levelno))



"""Tests for coBib's Database class."""

import copy
import logging
import os
import tempfile
from pathlib import Path
from shutil import copyfile
from typing import Any, Generator, Tuple

import pytest

from cobib.config import LabelSuffix, config
from cobib.database import Database, Entry

from .. import get_resource

TMPDIR = Path(tempfile.gettempdir())
EXAMPLE_LITERATURE = get_resource("example_literature.yaml")

DUMMY_ENTRY = Entry(
    "dummy",
    {
        "ENTRYTYPE": "misc",
        "author": "D. Dummy",
        "title": "Something dumb",
    },
)

DUMMY_ENTRY_YAML = """---
dummy:
    ENTRYTYPE: misc
    author: D. Dummy
    title: Something dumb
...
"""


@pytest.fixture(autouse=True)
def setup() -> Generator[Any, None, None]:
    """Setup debugging configuration.

    This method also clears the `Database` after each test run.
    It is automatically enabled for all tests in this file.

    Yields:
        Access to the local fixture variables.
    """
    config.load(get_resource("debug.py"))
    yield
    Database().clear()
    Database().read()
    config.defaults()


def test_database_singleton() -> None:
    """Test the Database is a Singleton."""
    bib = Database()
    bib2 = Database()
    assert bib is bib2


def test_database_missing_file(caplog: pytest.LogCaptureFixture) -> None:
    """Test exit upon missing database file.

    Args:
        caplog: the built-in pytest fixture.
    """
    config.database.file = TMPDIR / "cobib_test_missing_file.yaml"
    try:
        with pytest.raises(SystemExit):
            Database().read()
        for source, level, message in caplog.record_tuples:
            if ("cobib.database.database", logging.CRITICAL) == (
                source,
                level,
            ) and f"The database file {config.database.file} does not exist!" in message:
                break
        else:
            pytest.fail("coBib did not exit upon encountering a missing database file!")
    finally:
        config.database.file = EXAMPLE_LITERATURE


def test_database_update() -> None:
    """Test the `cobib.database.Database.update` method."""
    entries = {"dummy1": "test1", "dummy2": "test2", "dummy3": "test3"}
    bib = Database()
    bib.update(entries)  # type: ignore
    # pylint: disable=protected-access
    assert Database._unsaved_entries == {e: e for e in list(entries.keys())}
    for key, val in entries.items():
        assert bib[key] == val


def test_database_pop() -> None:
    """Test the `cobib.database.Database.pop` method."""
    entries = {"dummy1": "test1", "dummy2": "test2", "dummy3": "test3"}
    bib = Database()
    bib.update(entries)  # type: ignore
    # pylint: disable=protected-access
    Database._unsaved_entries = {}
    assert Database._unsaved_entries == {}  # pylint: disable=C1803
    entry = bib.pop("dummy1")
    assert entry == "test1"
    assert "dummy1" not in bib.keys()
    assert Database._unsaved_entries == {"dummy1": None}


def test_database_rename() -> None:
    """Test the `cobib.database.Database.rename` method."""
    bib = Database()
    # pylint: disable=protected-access
    Database._unsaved_entries = {}
    assert Database._unsaved_entries == {}  # pylint: disable=C1803
    bib.rename("einstein", "dummy")
    # pylint: disable=protected-access
    assert Database._unsaved_entries == {"einstein": "dummy"}


@pytest.mark.parametrize(
    ["label_suffix", "expected"],
    [
        [("_", LabelSuffix.ALPHA), "test_a"],
        [("_", LabelSuffix.CAPTIAL), "test_A"],
        [("_", LabelSuffix.NUMERIC), "test_1"],
        [(".", LabelSuffix.ALPHA), "test.a"],
        [(".", LabelSuffix.CAPTIAL), "test.A"],
        [(".", LabelSuffix.NUMERIC), "test.1"],
    ],
)
def test_database_disambiguate_label(label_suffix: Tuple[str, LabelSuffix], expected: str) -> None:
    # pylint: disable=invalid-name
    """Test the `cobib.database.Database.disambiguate_label` method."""
    config.database.format.label_default = "test"
    config.database.format.label_suffix = label_suffix

    entries = {"dummy": "test", "test": "no"}
    bib = Database()
    bib.update(entries)  # type: ignore

    new_label = Database().disambiguate_label("test", entries["dummy"])  # type: ignore
    assert new_label == expected


def test_database_read() -> None:
    """Test the `cobib.database.Database.read` method."""
    bib = Database()
    bib.read()
    # pylint: disable=protected-access
    assert Database._unsaved_entries == {}  # pylint: disable=C1803
    assert list(bib.keys()) == ["einstein", "latexcompanion", "knuthwebsite"]


def test_database_save_add() -> None:
    """Test the `cobib.database.Database.save` method after entry addition."""
    # prepare temporary database
    config.database.file = TMPDIR / "cobib_test_database_file.yaml"
    copyfile(EXAMPLE_LITERATURE, config.database.file)

    # initialize database
    bib = Database()
    bib.read()
    bib.update({"dummy": DUMMY_ENTRY})
    bib.save()

    expected = []
    with open(EXAMPLE_LITERATURE, "r", encoding="utf-8") as file:
        expected.extend(file.readlines())
    expected.extend(DUMMY_ENTRY_YAML.split("\n"))

    try:
        # pylint: disable=protected-access
        assert Database._unsaved_entries == {}  # pylint: disable=C1803

        with open(config.database.file, "r", encoding="utf-8") as file:
            # NOTE: do NOT use zip_longest to omit last entries (for testing simplicity)
            for line, truth in zip(file, expected):
                assert line.strip() == truth.strip()
            with pytest.raises(StopIteration):
                next(file)
    finally:
        os.remove(config.database.file)
        config.database.file = EXAMPLE_LITERATURE


def test_database_save_modify() -> None:
    """Test the `cobib.database.Database.save` method after entry modification."""
    # prepare temporary database
    config.database.file = TMPDIR / "cobib_test_database_file.yaml"
    copyfile(EXAMPLE_LITERATURE, config.database.file)

    # initialize database
    bib = Database()
    bib.read()
    entry = copy.deepcopy(bib["einstein"])
    entry.data["tags"] = "test"
    bib.update({"einstein": entry})
    bib.save()

    try:
        # pylint: disable=protected-access
        assert Database._unsaved_entries == {}  # pylint: disable=C1803

        with open(config.database.file, "r", encoding="utf-8") as file:
            with open(EXAMPLE_LITERATURE, "r", encoding="utf-8") as expected:
                # NOTE: do NOT use zip_longest to omit last entries (for testing simplicity)
                for line, truth in zip(file, expected):
                    if "tags" in line:
                        assert line == "  tags: test\n"
                        # advance only the `file` iterator one step in order to catch up with the
                        # `expected` iterator
                        line = next(file)
                    assert line == truth
                with pytest.raises(StopIteration):
                    next(file)
    finally:
        os.remove(config.database.file)
        config.database.file = EXAMPLE_LITERATURE


def test_database_save_delete() -> None:
    """Test the `cobib.database.Database.save` method after entry deletion."""
    # prepare temporary database
    config.database.file = TMPDIR / "cobib_test_database_file.yaml"
    copyfile(EXAMPLE_LITERATURE, config.database.file)

    # initialize database
    bib = Database()
    bib.read()
    bib.pop("knuthwebsite")
    bib.save()

    try:
        # pylint: disable=protected-access
        assert Database._unsaved_entries == {}  # pylint: disable=C1803

        with open(config.database.file, "r", encoding="utf-8") as file:
            with open(EXAMPLE_LITERATURE, "r", encoding="utf-8") as expected:
                # NOTE: do NOT use zip_longest to omit last entries (for testing simplicity)
                for line, truth in zip(file, expected):
                    assert line == truth
                with pytest.raises(StopIteration):
                    next(file)
    finally:
        os.remove(config.database.file)
        config.database.file = EXAMPLE_LITERATURE



"""Tests for coBib's AddCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import logging
import os
import tempfile
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Type

import pytest
from typing_extensions import override

from cobib.commands import AddCommand
from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .. import MockStdin, get_resource
from .command_test import CommandTest

EXAMPLE_LITERATURE = get_resource("example_literature.yaml")
EXAMPLE_DUPLICATE_ENTRY_BIB = get_resource("example_duplicate_entry.bib", "commands")
EXAMPLE_DUPLICATE_ENTRY_YAML = get_resource("example_duplicate_entry.yaml", "commands")
EXAMPLE_MULTI_FILE_ENTRY_BIB = get_resource("example_multi_file_entry.bib", "commands")
EXAMPLE_MULTI_FILE_ENTRY_YAML = get_resource("example_multi_file_entry.yaml", "commands")

if TYPE_CHECKING:
    import _pytest.fixtures

    import cobib.commands


class TestAddCommand(CommandTest):
    """Tests for coBib's AddCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return AddCommand

    @pytest.fixture
    def post_setup(
        self, monkeypatch: pytest.MonkeyPatch, request: _pytest.fixtures.SubRequest
    ) -> Generator[Dict[str, Any], None, None]:
        """Additional setup instructions.

        Args:
            monkeypatch: the built-in pytest fixture.
            request: a pytest sub-request providing access to nested parameters.

        Yields:
            The internally used parameters for potential later re-use during the actual test.
        """
        if not hasattr(request, "param"):
            # use default settings
            request.param = {"stdin_list": None}

        monkeypatch.setattr("sys.stdin", MockStdin(request.param.get("stdin_list", None)))

        yield request.param

    def _assert(self, extra_filename: str) -> None:
        """Common assertion utility method.

        Args:
            extra_filename: path to an additional filename whose contents are to be added to the
                expected lines.
        """
        # compare with reference file
        with open(EXAMPLE_LITERATURE, "r", encoding="utf-8") as expected:
            true_lines = expected.readlines()
        with open(extra_filename, "r", encoding="utf-8") as extra:
            true_lines += extra.readlines()
        with open(config.database.file, "r", encoding="utf-8") as file:
            # we use zip_longest to ensure that we don't have more than we expect
            for line, truth in zip_longest(file, true_lines):
                assert line == truth

    def _assert_entry(self, label: str, **kwargs) -> None:  # type: ignore
        """An additional assertion utility to check specific entry fields.

        Args:
            label: the label of the entry.
            kwargs: additional keyword arguments whose contents are checked against the Entry's
                `data contents.
        """
        entry = Database()[label]
        for key, value in kwargs.items():
            assert entry.data.get(key, None) == value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["more_args", "entry_kwargs"],
        [
            [[], {}],
            [
                ["-f", "test/debug.py"],
                {"file": [str(RelPath("test/debug.py"))]},
            ],
            [["-l", "dummy_label"], {}],
            [["tag"], {"tags": ["tag"]}],
            [["tag", "tag2"], {"tags": ["tag", "tag2"]}],
        ],
    )
    async def test_command(
        self, setup: Any, more_args: List[str], entry_kwargs: Dict[str, Any]
    ) -> None:
        # pylint: disable=invalid-overridden-method
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            more_args: additional arguments to be passed to the command.
            entry_kwargs: the expected contents of the resulting `Entry`.
        """
        git = setup.get("git", False)

        try:
            label = more_args[more_args.index("-l") + 1]
        except ValueError:
            label = "example_multi_file_entry"
        args = ["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB] + more_args

        await AddCommand(*args).execute()

        assert Database()[label]

        if entry_kwargs or label != "example_multi_file_entry":
            self._assert_entry(label, **entry_kwargs)
        else:
            # only when we don't use extra arguments the files will match
            self._assert(EXAMPLE_MULTI_FILE_ENTRY_YAML)

        if git:
            # assert the git commit message
            # Note: we do not assert the arguments, because they depend on the available parsers
            self.assert_git_commit_message("add", None)

    @pytest.mark.asyncio
    async def test_add_new_entry(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test adding a new plain entry.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await AddCommand("-l", "dummy").execute()
        assert (
            "cobib.commands.add",
            30,
            "No input to parse. Creating new entry 'dummy' manually.",
        ) in caplog.record_tuples

        with open(config.database.file, "r", encoding="utf-8") as file:
            lines = file.readlines()
            dummy_start = lines.index("dummy:\n")
            assert dummy_start > 0
            assert lines[dummy_start - 1] == "---\n"
            assert lines[dummy_start + 1] == "  ENTRYTYPE: article\n"
            assert lines[dummy_start + 2] == "...\n"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("folder", [None, "."])
    @pytest.mark.parametrize("skip_download", [None, True, False])
    @pytest.mark.parametrize("config_overwrite", [True, False])
    async def test_add_with_download(
        self,
        setup: Any,
        folder: Optional[str],
        skip_download: Optional[bool],
        config_overwrite: bool,
        capsys: pytest.CaptureFixture[str],
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test adding a new entry with possibly an associated file automatically downloaded.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            folder: the folder for the downloaded file.
            skip_download: argument to `AddCommand`.
            config_overwrite: what to overwrite `config.commands.add.skip_download` with.
            capsys: the built-in pytest fixture.
            caplog: the built-in pytest fixture.
        """
        config.commands.add.skip_download = config_overwrite

        should_download = not config_overwrite
        if skip_download is not None:
            should_download = not skip_download

        path = RelPath(f"{'/tmp' if folder is None else folder}/Cao2018.pdf")
        try:
            # ensure file does not exist yet
            os.remove(path.path)
        except FileNotFoundError:
            pass
        try:
            args = ["-a", "1812.09976"]
            if folder:
                args += ["-p", folder]
            if skip_download is not None:
                args.append(f"--{'skip' if skip_download else 'force'}-download")

            await AddCommand(*args).execute()

            if (
                "cobib.parsers.arxiv",
                logging.ERROR,
                "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
            ) in caplog.record_tuples:
                pytest.skip("The requests API encountered an error. Skipping test.")

            entry = Database()["Cao2018"]
            assert entry.label == "Cao2018"
            assert entry.data["archivePrefix"] == "arXiv"
            assert entry.data["arxivid"].startswith("1812.09976")
            assert "_download" not in entry.data.keys()

            if should_download:
                assert f"Successfully downloaded {path}" in capsys.readouterr().out
                assert os.path.exists(path.path)
            else:
                assert not os.path.exists(path.path)
        finally:
            try:
                os.remove(path.path)
            except FileNotFoundError:
                pass

    @pytest.mark.asyncio
    @pytest.mark.parametrize("deprecated", (True, False))
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    async def test_add_with_update(
        self, setup: Any, deprecated: bool, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test update option of AddCommand.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            deprecated: whether to test the deprecated or new means.
            caplog: the built-in pytest fixture.
        """
        git = setup.get("git", False)
        await AddCommand("-a", "1812.09976", "--skip-download").execute()

        if (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        # assert initial state
        entry = Database()["Cao2018"]

        assert entry.data["author"].startswith("Yudong Cao")
        assert entry.data["title"].startswith("Quantum Chemistry in the Age of Quantum Computing")
        assert entry.data["arxivid"].startswith("1812.09976")
        assert entry.data["doi"] == "10.1021/acs.chemrev.8b00803"
        assert entry.data["primaryClass"] == "quant-ph"
        assert entry.data["archivePrefix"] == "arXiv"
        assert entry.data["abstract"] != ""
        assert entry.data["year"] == 2018

        assert "journal" not in entry.data.keys()
        assert "month" not in entry.data.keys()
        assert "number" not in entry.data.keys()
        assert "pages" not in entry.data.keys()
        assert "volume" not in entry.data.keys()

        args = [
            "-d",
            "10.1021/acs.chemrev.8b00803",
            "-l",
            "Cao2018",
            "--skip-download",
        ]
        if deprecated:
            args += ["--update"]
        else:
            args += ["--disambiguation", "update"]
        await AddCommand(*args).execute()

        if (
            "cobib.parsers.doi",
            logging.ERROR,
            "An Exception occurred while trying to query the DOI: 10.1021/acs.chemrev.8b00803.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        if deprecated:
            assert (
                "cobib.commands.add",
                logging.WARNING,
                "The '--update' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation update'.",
            ) in caplog.record_tuples

        # assert final state
        entry = Database()["Cao2018"]

        assert entry.data["author"].startswith("Yudong Cao")
        assert entry.data["title"].startswith("Quantum Chemistry in the Age of Quantum Computing")
        assert entry.data["arxivid"].startswith("1812.09976")
        assert entry.data["primaryClass"] == "quant-ph"
        assert entry.data["archivePrefix"] == "arXiv"
        assert entry.data["abstract"] != ""

        assert entry.data["journal"] == "Chemical Reviews"
        assert entry.data["doi"] == "10.1021/acs.chemrev.8b00803"
        assert entry.data["month"] == "aug"
        assert entry.data["number"] == 19
        assert entry.data["pages"] == "10856--10915"
        assert entry.data["volume"] == 119
        assert entry.data["year"] == 2019

        if git:
            # assert the git commit message
            # Note: we do not assert the arguments, because they depend on the available parsers
            self.assert_git_commit_message("add", None)

    @pytest.mark.asyncio
    async def test_skip_manual_add_if_exists(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test manual addition is skipped if the label exists already.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await AddCommand("-l", "einstein").execute()
        assert (
            "cobib.commands.add",
            30,
            (
                "You tried to add the 'einstein' entry manually, but it already exists, "
                "please use `cobib edit einstein` instead!"
            ),
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize("deprecated", (True, False))
    async def test_continue_after_skip_exists(
        self, setup: Any, deprecated: bool, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test entry addition continues after skipping over existing entry.

        Regression test against #83

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            deprecated: whether to test the deprecated or new means.
            caplog: the built-in pytest fixture.
        """
        with tempfile.NamedTemporaryFile("w") as file:
            with open(EXAMPLE_DUPLICATE_ENTRY_BIB, "r", encoding="utf-8") as existing:
                file.writelines(existing.readlines())
            file.writelines(["@article{dummy,\nauthor = {Dummy},\n}"])
            file.flush()
            args = ["-b", file.name]
            if deprecated:
                args += ["--skip-existing"]
            else:
                args += ["--disambiguation", "keep"]
            await AddCommand(*args).execute()
        assert (
            "cobib.commands.add",
            20,
            "Skipping addition of the already existing entry 'einstein'.",
        ) in caplog.record_tuples
        assert (
            "cobib.database.database",
            10,
            "Updating entry dummy",
        ) in caplog.record_tuples

        if deprecated:
            assert (
                "cobib.commands.add",
                logging.WARNING,
                "The '--skip-existing' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation keep'.",
            ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_warning_missing_label(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test warning for missing label and any other input.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await AddCommand().execute()
        assert (
            "cobib.commands.add",
            40,
            "Neither an input to parse nor a label for manual creation specified!",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    async def test_overwrite_label(self, setup: Any) -> None:
        """Test add command while specifying a label manually.

        Regression test against #4.

        The duplicate entry has been adapted to also assert the elongation of Journal names.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        config.utils.journal_abbreviations = [("Annalen der Physik", "Ann. Phys.")]
        git = setup.get("git", False)
        # add potentially duplicate entry
        await AddCommand(
            "-b", EXAMPLE_DUPLICATE_ENTRY_BIB, "--label", "duplicate_resolver"
        ).execute()

        assert Database()["duplicate_resolver"]

        self._assert(EXAMPLE_DUPLICATE_ENTRY_YAML)

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    async def test_configured_label_default(self, setup: Any) -> None:
        """Test add command when a `label_default` is pre-configured.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        config.database.format.label_default = "{author.split()[1]}{year}"
        git = setup.get("git", False)

        await AddCommand("-b", EXAMPLE_DUPLICATE_ENTRY_BIB).execute()

        assert Database()["Einstein1905"]

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup", "post_setup", "args"],
        [
            [{"git": False}, {"stdin_list": ["disambiguate"]}, []],
            [{"git": True}, {"stdin_list": ["disambiguate"]}, []],
            [{"git": False}, {}, ["--disambiguation", "disambiguate"]],
            [{"git": True}, {}, ["--disambiguation", "disambiguate"]],
        ],
        indirect=["setup", "post_setup"],
    )
    async def test_disambiguate_label(
        self, setup: Any, post_setup: Any, args: list[str], caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test label disambiguation if the provided one already exists.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            post_setup: an additional setup fixture.
            args: additional arguments for the AddCommand.
            caplog: the built-in pytest fixture.
        """
        git = setup.get("git", False)

        await AddCommand("-b", EXAMPLE_DUPLICATE_ENTRY_BIB, *args).execute()

        assert (
            "cobib.commands.add",
            30,
            "You tried to add a new entry 'einstein' which already exists!",
        ) in caplog.record_tuples
        assert (
            "cobib.commands.add",
            30,
            "The label will be disambiguated based on the configuration option: "
            "config.database.format.label_suffix",
        ) in caplog.record_tuples

        assert Database()["einstein_a"]

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["post_setup"],
        [
            [{"stdin_list": ["disambiguate"]}],
        ],
        indirect=["post_setup"],
    )
    async def test_disambiguate_download(
        self,
        setup: Any,
        post_setup: Any,
        capsys: pytest.CaptureFixture[str],
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test label disambiguation is propagated to downloaded files.

        This is a regression test against https://gitlab.com/cobib/cobib/-/issues/96.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            post_setup: an additional setup fixture.
            capsys: the built-in pytest fixture.
            caplog: the built-in pytest fixture.
        """
        database = Database()

        for label in ("Cao2018", "Cao2018_a"):
            try:
                path = RelPath(f"/tmp/{label}.pdf")
                try:
                    # ensure file does not exist yet
                    os.remove(path.path)
                except FileNotFoundError:
                    pass

                # by repeatedly calling the same add command, we trigger the label disambiguation
                await AddCommand("-a", "1812.09976").execute()

                if (
                    "cobib.parsers.arxiv",
                    logging.ERROR,
                    "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
                ) in caplog.record_tuples:
                    pytest.skip("The requests API encountered an error. Skipping test.")

                entry = database[label]
                assert entry.data["author"].startswith("Yudong Cao")
                assert entry.data["title"].startswith(
                    "Quantum Chemistry in the Age of Quantum Computing"
                )
                assert entry.data["arxivid"].startswith("1812.09976")
                assert entry.data["doi"] == "10.1021/acs.chemrev.8b00803"
                assert entry.data["primaryClass"] == "quant-ph"
                assert entry.data["archivePrefix"] == "arXiv"
                assert entry.data["abstract"] != ""
                assert entry.data["year"] == 2018
                assert os.path.exists(path.path)

            finally:
                try:
                    os.remove(path.path)
                except FileNotFoundError:
                    pass

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
        ],
        indirect=["setup"],
    )
    # other variants are already covered by test_command
    async def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
        """
        await self.run_module(
            monkeypatch, "main", ["cobib", "add", "-b", EXAMPLE_MULTI_FILE_ENTRY_BIB]
        )
        self._assert(EXAMPLE_MULTI_FILE_ENTRY_YAML)

    @pytest.mark.asyncio
    async def test_event_pre_add_command(self, setup: Any) -> None:
        """Tests the PreAddCommand event."""

        @Event.PreAddCommand.subscribe
        def hook(command: AddCommand) -> None:
            command.largs.label = "dummy"

        assert Event.PreAddCommand.validate()

        await AddCommand("-b", EXAMPLE_DUPLICATE_ENTRY_BIB).execute()

        assert "dummy" in Database().keys()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["post_setup"],
        [
            [{"stdin_list": ["disambiguate"]}],
        ],
        indirect=["post_setup"],
    )
    async def test_event_post_add_command(self, setup: Any, post_setup: Any) -> None:
        """Tests the PostAddCommand event."""

        @Event.PostAddCommand.subscribe
        def hook(command: AddCommand) -> None:
            command.new_entries["dummy"] = command.new_entries.pop("einstein_a")

        assert Event.PostAddCommand.validate()

        await AddCommand("-b", EXAMPLE_DUPLICATE_ENTRY_BIB).execute()

        assert "dummy" in Database().keys()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        [
            "args",
            "msg_string",
        ],
        [
            [
                ["--skip-existing"],
                "The '--skip-existing' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation keep'.",
            ],
            [
                ["--update"],
                "The '--update' argument of the 'add' command is deprecated! "
                "Instead you should use '--disambiguation update'.",
            ],
        ],
    )
    async def test_warn_deprecated_args(
        self, setup: Any, args: list[str], msg_string: str, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test the warning upon usage of deprecated arguments.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: arguments for the AddCommand.
            msg_string: the string which the warning message must match.
            caplog: the built-in pytest fixture.
        """
        args = ["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB] + args

        await AddCommand(*args).execute()

        for source, level, msg in caplog.record_tuples:
            if source == "cobib.commands.add" and level == 30:
                if msg == msg_string:
                    break
        else:
            pytest.fail("No Warning logged from AddCommand.")



"""Tests for coBib's main executable.

Since all commands are tested separately in their respective test files, the tests here mainly deal
with testing the global parser arguments.
"""
# pylint: disable=unused-argument

import logging
import os
import tempfile
from pathlib import Path
from typing import Any, List

import pytest

from cobib.config import config

from . import get_resource
from .cmdline_test import CmdLineTest


class TestMainExecutable(CmdLineTest):
    """Tests for coBib's main executable."""

    @staticmethod
    @pytest.fixture
    def setup() -> None:
        """Load testing config."""
        config.load(get_resource("debug.py"))

    @pytest.mark.asyncio
    async def test_version(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Tests the version parser argument.

        Args:
            setup: a local pytest fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        with pytest.raises(SystemExit):
            await self.run_module(monkeypatch, "main", ["cobib", "--version"])
        # pylint: disable=import-outside-toplevel
        from cobib import __version__

        assert capsys.readouterr().out.strip() == f"coBib v{__version__}"

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "args",
        [
            ["open", "einstein"],
            ["_example_config"],
        ],
    )
    @pytest.mark.parametrize(
        ["verbosity_arg", "level"],
        [
            ["-v", logging.INFO],
            ["-vv", logging.DEBUG],
        ],
    )
    async def test_verbosity(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        args: List[str],
        verbosity_arg: str,
        level: int,
    ) -> None:
        """Tests the verbosity parser argument.

        Args:
            setup: a local pytest fixture.
            monkeypatch: the built-in pytest fixture.
            args: the list of values with which to monkeypatch `sys.argv`.
            verbosity_arg: the value of the verbosity argument.
            level: the level of the verbosity argument.
        """
        # we choose the open command as an arbitrary choice which has minimal side effects
        args = ["cobib"] + args
        if verbosity_arg:
            args.insert(1, verbosity_arg)
        await self.run_module(monkeypatch, "main", args)
        assert logging.getLogger().getEffectiveLevel() == logging.DEBUG
        assert logging.getLogger().handlers[-1].level == level

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "args",
        [
            ["open", "einstein"],
            ["_example_config"],
        ],
    )
    async def test_logfile(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, args: List[str]
    ) -> None:
        """Tests the logfile parser argument.

        Args:
            setup: a local pytest fixture.
            monkeypatch: the built-in pytest fixture.
            args: the list of values with which to monkeypatch `sys.argv`.
        """
        logfile = str(Path(tempfile.gettempdir()) / "cobib_test_logging.log")
        # we choose the open command as an arbitrary choice which has minimal side effects
        await self.run_module(monkeypatch, "main", ["cobib", "-l", logfile] + args)
        try:
            assert isinstance(logging.getLogger().handlers[-1], logging.FileHandler)
            assert logging.getLogger().handlers[-1].baseFilename == logfile  # type: ignore
        finally:
            os.remove(logfile)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "args",
        [
            ["open", "einstein"],
            ["_example_config"],
        ],
    )
    async def test_configfile(self, monkeypatch: pytest.MonkeyPatch, args: List[str]) -> None:
        """Tests the configfile parser argument.

        Args:
            monkeypatch: the built-in pytest fixture.
            args: the list of values with which to monkeypatch `sys.argv`.
        """
        # we choose the open command as an arbitrary choice which has minimal side effects
        await self.run_module(monkeypatch, "main", ["cobib", "-c", get_resource("debug.py")] + args)
        assert config.database.file == get_resource("example_literature.yaml")



"""Tests for coBib's URLParser."""
# pylint: disable=unused-argument

import logging
from typing import Callable, Dict, Optional

import pytest

from cobib.config import Event
from cobib.database import Entry
from cobib.parsers import URLParser

from .parser_test import ParserTest
from .test_arxiv import assert_default_test_entry as assert_arxiv_entry
from .test_doi import assert_default_test_entry as assert_doi_entry


def assert_default_test_entry(entry: Entry) -> None:
    """Asserts that the passed entry is the default testing entry.

    Args:
        entry: the entry to assert.
    """
    entry.escape_special_chars()
    assert entry.label == "Grimsley_2019"
    assert entry.data["doi"] == "10.1038/s41467-019-10988-2"
    assert entry.data["url"] == ["https://doi.org/10.1038%2Fs41467-019-10988-2"]
    assert entry.data["year"] == 2019
    assert entry.data["month"] == "jul"
    assert entry.data["publisher"] == "Springer Science and Business Media {LLC}"
    assert entry.data["volume"] == 10
    assert entry.data["number"] == 1
    assert (
        entry.data["author"]
        == "Harper R. Grimsley and Sophia E. Economou and Edwin Barnes and Nicholas J. Mayhall"
    )
    assert (
        entry.data["title"]
        == "An adaptive variational algorithm for exact molecular simulations on a quantum computer"
    )
    assert entry.data["journal"] == "Nature Communications"


class TestURLParser(ParserTest):
    """Tests for coBib's URLParser."""

    @pytest.mark.parametrize(
        ("query", "assertion"),
        [
            ("https://arxiv.org/abs/1812.09976", assert_arxiv_entry),
            ("https://doi.org/10.1021/acs.chemrev.8b00803", assert_doi_entry),
            ("https://www.nature.com/articles/s41467-019-10988-2", assert_default_test_entry),
        ],
    )
    def test_from_url(
        self, query: str, assertion: Callable[[Entry], None], caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test parsing from URL.

        Args:
            query: the URL which to query.
            assertion: the assertion method to run.
            caplog: the built-in pytest fixture.
        """
        entries = URLParser().parse(query)

        try:
            entry = list(entries.values())[0]
        except IndexError:
            pytest.skip("Skipping because we likely ran into a network timeout.")

        assertion(entry)

    def test_invalid_url(self) -> None:
        """Test parsing an invalid URL."""
        entries = URLParser().parse("https://github.com/")
        assert not entries
        assert entries == {}

    def test_dump(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test dumping.

        Args:
            caplog: the built-in pytest fixture.
        """
        entry = Entry("dummy", {"ENTRYTYPE": "unpublished"})
        URLParser().dump(entry)

        assert (
            "cobib.parsers.url",
            logging.ERROR,
            "Cannot dump an entry as a URL.",
        ) in caplog.record_tuples

    def test_event_pre_url_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PreURLParse event."""

        @Event.PreURLParse.subscribe
        def hook(string: str) -> Optional[str]:
            return "https://www.nature.com/articles/s41467-019-10988-2"

        assert Event.PreURLParse.validate()

        entries = URLParser().parse("Hello world!")
        if any(s == "cobib.parsers.url" and t == logging.ERROR for s, t, _ in caplog.record_tuples):
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    def test_event_post_url_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PostURLParse event."""

        @Event.PostURLParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Grimsley_2019"].data["test"] = "dummy"

        assert Event.PostURLParse.validate()

        entries = URLParser().parse("https://www.nature.com/articles/s41467-019-10988-2")
        if any(s == "cobib.parsers.url" and t == logging.ERROR for s, t, _ in caplog.record_tuples):
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)
        assert entry.data["test"] == "dummy"



"""coBib's configuration module.

The configuration has undergone a major revision during v3.0 when it was redesigned from loading a
`INI` file with the `configparser` module to become a standalone Python object.
The configuration will be read from the file that appears first in the following list:

1. a file pointed to by the `--config` command-line option.
2. the file located at the path provided by the environment variable `COBIB_CONFIG`. If this
   variable is set to one of the values (`"", 0, "f", "false", "nil", "none"`), no configuration
   file will be loaded and the rest of this list is skipped.
3. the XDG-default location: `~/.config/cobib/config.py`.
4. if none of the above are found, coBib will fallback to the default configuration.

A configuration file is *not* required to set all attributes itself.
In fact, it is required to import the default configuration first and then simply overwrite any
settings to the user's liking.

```python
from cobib.config import config
```

For more information take a look at the example configuration, `cobib.config.example`.
"""

from .config import LabelSuffix as LabelSuffix
from .config import config as config
from .event import Event as Event



"""Tests for coBib's BibtexParser."""
# pylint: disable=unused-argument

from typing import Dict, Optional

import pytest

from cobib.config import Event
from cobib.database import Entry
from cobib.parsers import BibtexParser

from .parser_test import ParserTest


class TestBibtexParser(ParserTest):
    """Tests for coBib's BibtexParser."""

    def test_to_bibtex(self) -> None:
        """Test to bibtex conversion."""
        pytest.skip("Testing this string is a bit ambigious. Assumed to be tested by bibtexparser.")

    def test_from_bibtex_str(self) -> None:
        """Test parsing a bibtex string."""
        reference = self.EXAMPLE_ENTRY_DICT.copy()
        with open(self.EXAMPLE_BIBTEX_FILE, "r", encoding="utf-8") as file:
            bibtex_str = file.read()
        entries = BibtexParser().parse(bibtex_str)
        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_from_bibtex_file(self) -> None:
        """Test parsing a bibtex file."""
        reference = self.EXAMPLE_ENTRY_DICT.copy()
        entries = BibtexParser().parse(self.EXAMPLE_BIBTEX_FILE)
        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_event_pre_bibtex_parse(self) -> None:
        """Tests the PreBibtexParse event."""

        @Event.PreBibtexParse.subscribe
        def hook(string: str) -> Optional[str]:
            with open(self.EXAMPLE_BIBTEX_FILE, "r", encoding="utf-8") as file:
                return file.read()

        assert Event.PreBibtexParse.validate()

        reference = self.EXAMPLE_ENTRY_DICT.copy()
        entries = BibtexParser().parse("Hello world!")

        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_event_post_bibtex_parse(self) -> None:
        """Tests the PostBibtexParse event."""

        @Event.PostBibtexParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Cao_2019"].data["month"] = "August"

        assert Event.PostBibtexParse.validate()

        reference = self.EXAMPLE_ENTRY_DICT.copy()
        reference["month"] = "August"

        with open(self.EXAMPLE_BIBTEX_FILE, "r", encoding="utf-8") as file:
            bibtex_str = file.read()
        entries = BibtexParser().parse(bibtex_str)
        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_event_pre_bibtex_dump(self) -> None:
        """Tests the PreBibtexDump event."""

        @Event.PreBibtexDump.subscribe
        def hook(entry: Entry) -> None:
            entry.label = "Cao2019"

        assert Event.PreBibtexDump.validate()

        entry = Entry("Cao_2019", self.EXAMPLE_ENTRY_DICT.copy())
        entry_str = BibtexParser().dump(entry)
        assert entry_str.split("\n")[0] == "@article{Cao2019,"

    def test_event_post_bibtex_dump(self) -> None:
        """Tests the PostBibtexDump event."""

        @Event.PostBibtexDump.subscribe
        def hook(string: str) -> Optional[str]:
            return "Hello world!"

        assert Event.PostBibtexDump.validate()

        entry = Entry("Cao_2019", self.EXAMPLE_ENTRY_DICT.copy())
        entry_str = BibtexParser().dump(entry)
        assert entry_str == "Hello world!"



"""coBib's utilities.

This module provides some general purpose utilities.
This includes both, general purpose objects used throughout coBib's source code and test suite, as
well as helper methods exposed via the command line interface.
"""



"""An example configuration for coBib.

Since version 3.0 coBib is configured through a Python file.
For documentation purposes this example configuration lists all possible settings with their default
values and detailed explanations.

Internally, coBib's configuration is nothing but a (nested) Python dataclass. This means, you can
simply change any setting via an attribute like so:

```python
config.database.git = True
```
"""

# Generally, you won't need these, but the default configuration relies on them.
import os
import sys

# To get started you must import coBib's configuration.
from cobib.config import LabelSuffix, config

# Now, you are all set to apply your own settings.


# LOGGING
# You can specify the default cache location.
config.logging.cache = "~/.cache/cobib/cache"
# You can specify the default logfile location.
config.logging.logfile = "~/.cache/cobib/cobib.log"
# You can also set the location for the cached version number based on which coBib shows you the
# latest changes. You can set this to `None` to disable this functionality entirely.
config.logging.version = "~/.cache/cobib/version"

# COMMANDS
# These settings affect some command specific behavior.

# You can specify whether the automatic file download should be skipped during entry addition.
config.commands.add.skip_download = False

# You can specify whether you should be prompted for confirmation before deleting an entry.
config.commands.delete.confirm = True
# You can specify whether associated files should be preserved during entry deletion.
config.commands.delete.preserve_files = False

# You can specify the default bibtex entry type.
config.commands.edit.default_entry_type = "article"
# You can specify the editor program. Note, that this default will respect your `$EDITOR`
# environment setting and fall back to `vim` if that variable is not set.
config.commands.edit.editor = os.environ.get("EDITOR", "vim")
# You can specify whether associated files should be preserved when renaming during editing.
config.commands.edit.preserve_files = False

# You can specify whether downloading of attachments inside the imported library should be skipped.
config.commands.import_.skip_download = False

# You can configure the default columns displayed during the list command.
config.commands.list_.default_columns = ["label", "title"]
# You can specify whether filter matching should be performed case-insensitive.
config.commands.list_.ignore_case = False

# You can specify whether associated files should be preserved when renaming during modifying.
config.commands.modify.preserve_files = False

# You can specify a custom command which will be used to `open` files associated with your entries.
config.commands.open.command = "xdg-open" if sys.platform.lower() == "linux" else "open"
# You can specify the names of the data fields which are to be checked for openable URLs.
config.commands.open.fields = ["file", "url"]

# You can specify the default number of context lines to be provided for each search query match.
# This is similar to the `-C` option of `grep`.
config.commands.search.context = 1
# You can specify a custom grep tool which will be used to search through your database and any
# associated files. The default tool (`grep`) will not provide results for attached PDFs but other
# tools such as [ripgrep-all](https://github.com/phiresky/ripgrep-all) will.
config.commands.search.grep = "grep"
# If you want to specify additional arguments for your grep command, you can specify them as a list
# of strings in the following setting. Note, that GNU's grep understands extended regex patterns
# even without specifying `-E`.
config.commands.search.grep_args = []

# You can specify whether searches should be performed case-insensitive.
config.commands.search.ignore_case = False

# You can configure the search label and query highlights.
config.commands.search.highlights.label = "blue"
config.commands.search.highlights.query = "red"


# DATABASE
# These settings affect the database in general.

# You can specify the path to the database YAML file. You can use a `~` to represent your `$HOME`
# directory.
config.database.file = "~/.local/share/cobib/literature.yaml"

# coBib can integrate with `git` in order to automatically track the history of your database.
# However, by default, this option is disabled. If you want to enable it, simply change the
# following setting to `True` and initialize your database with `cobib init --git`.
# Warning: Before enabling this setting you must ensure that you have set up git properly by setting
# your name and email address.
config.database.git = False

# DATABASE.FORMAT
# You can also specify some aspects about the format of the database.

# You can specify a default label format which will be used for the database entry keys. The format
# of this option follows the f-string like formatting of modifications (see also the documentation
# of the [ModifyCommand](https://cobib.gitlab.io/cobib/cobib/commands/modify.html)). The default
# configuration value leaves the label unchanged compared to the metadata provided by the source
# from which the entry gets added. A more useful example is
#     `"{author.split(' and ')[0].split()[-1]}{year}"`
# which takes the surname of the first author and immediately appends the publication year.
config.database.format.label_default = "{label}"

# You can specify the suffix format which is used to disambiguate labels if a conflict would occur.
# This option takes a tuple of length 2, where the first entry is the string separating the proposed
# label from the enumerator and the second one is one of the enumerators provided in the
# `config.LabelSuffix` object. The available enumerators are:
#   - ALPHA: a, b, ...
#   - CAPTIAL: A, B, ...
#   - NUMERIC: 1, 2, ...
config.database.format.label_suffix = ("_", LabelSuffix.ALPHA)

# You can specify whether latex warnings should not be ignored during the escaping of special
# characters. This is a simple option which gets passed on to the internally used `pylatexenc`
# library.
config.database.format.suppress_latex_warnings = True

# DATABASE.STRINGIFY
# You can customize the functions which convert non-string fields.

# Three fields are currently explicitly stored as lists internally. Upon conversion to the BibTeX
# format, these need to be converted to a basic string. In this process the entries of the list will
# be joined using the separators configured by the following settings.
config.database.stringify.list_separator.file = ", "
config.database.stringify.list_separator.tags = ", "
config.database.stringify.list_separator.url = ", "

# PARSERS
# These settings affect some parser specific behavior.

# You can specify whether the bibtex-parser should ignore non-standard bibtex entry types.
config.parsers.bibtex.ignore_non_standard_types = False

# You can specify that the C-based implementation of the YAML parser (called `LibYAML`) shall be
# used, *significantly* increasing the performance of the parsing. Note, that this requires manual
# installation of the C-based parser:
# https://yaml.readthedocs.io/en/latest/install.html#optional-requirements
config.parsers.yaml.use_c_lib_yaml = True

# UTILS

# You can specify the default download location for associated files.
config.utils.file_downloader.default_location = "~/.local/share/cobib"

# You can provide rules to map from a journal's landing page URL to its PDF URL. To do so, you must
# insert an entry into the following dictionary, with a regex-pattern matching the journal's landing
# page URL and a value being the PDF URL. E.g.:
#
#     config.utils.file_downloader.url_map[
#         r"(.+)://aip.scitation.org/doi/([^/]+)"
#     ] = r"\1://aip.scitation.org/doi/pdf/\2"
#
#     config.utils.file_downloader.url_map[
#         r"(.+)://quantum-journal.org/papers/([^/]+)"
#     ] = r"\1://quantum-journal.org/papers/\2/pdf/"
#
# Make sure to use raw Python strings to ensure proper backslash-escaping.
config.utils.file_downloader.url_map = {}

# You can specify a list of journal abbreviations. This list should be formatted as tuples of the
# form: `(full journal name, abbreviation)`. The abbreviation should include any necessary
# punctuation which can be excluded upon export (see also `cobib export --help`).
config.utils.journal_abbreviations = []

# EVENTS
# coBib allows you to register hooks on various events to further customize its behavior to your
# liking. Although these functions will be registered in the following dictionary, we recommend you
# to use the function-decorators as explained below.
config.events = {}
# To subscribe to a certain event do something similar to the following:
#
#     from os import system
#     from pathlib import Path
#     from cobib.config import Event
#
#     @Event.PostInitCommand.subscribe
#     def add_remote(root: Path, file: Path) -> None:
#         system(f"git -C {root} remote add origin https://github.com/user/repo")
#
# Note, that the typing is required for the config validation to pass!
# For more information refer to the
# [online documentation](https://cobib.gitlab.io/cobib/cobib/config/event.html).



"""coBib's ISBN parser.

This parser is capable of generating `cobib.database.Entry` instances from a given ISBN.
It gathers the BibTex-encoded data from the ISBN API and parses the raw json data.

Note, that the openlibrary API does not contain all ISBNs and potential server errors will be caught
by the parser.
In the future, I hope to make the API backend configurable.

Since v3.3.0 this parser also supports URLs from which an ISBN can be extracted directly.

The parser is registered under the `-i` and `--isbn` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import json
import logging
import re
from collections import OrderedDict
from typing import Dict

import requests
from typing_extensions import override

from cobib.config import Event
from cobib.database import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)

ISBN_URL = "https://openlibrary.org/api/books?bibkeys=ISBN:"
"""ISBN API URL taken from [here](https://openlibrary.org/dev/docs/api/books)."""
ISBN_REGEX = re.compile(
    r"(97[89]{1}(?:-?\d){10}|\d{9}[0-9X]{1}|[-0-9X]{10,16})", re.I | re.M | re.S
)
"""A regex pattern used to match valid ISBNs. Adapted from
[here](https://github.com/xlcnd/isbnlib)."""


class ISBNParser(Parser):
    """The ISBN Parser."""

    name = "isbn"

    @override
    def parse(self, string: str) -> Dict[str, Entry]:
        string = Event.PreISBNParse.fire(string) or string

        try:
            match = re.search(ISBN_REGEX, string)
            if match is None:
                raise AssertionError
        except AssertionError:
            msg = f"'{string}' is not a valid ISBN."
            LOGGER.warning(msg)
            return OrderedDict()
        isbn = match.group(1)
        LOGGER.info("Gathering BibTex data for ISBN: %s.", isbn)
        isbn_plain = "".join([i for i in isbn if i.isdigit()])
        try:
            page = requests.get(ISBN_URL + isbn_plain + "&jscmd=data&format=json", timeout=10)
        except requests.exceptions.RequestException as err:
            LOGGER.error("An Exception occurred while trying to query the ISBN: %s.", isbn)
            LOGGER.error(err)
            return OrderedDict()
        try:
            contents = dict(json.loads(page.content))
        except json.JSONDecodeError as err:
            LOGGER.error("An Exception occurred while parsing the query results: %s.", page.content)
            LOGGER.error(err)
            return OrderedDict()
        if not contents:
            msg = (
                f'No data was found for ISBN "{isbn}". If you think this is an error and '
                + "the openlibrary API should provide an entry, please file a bug report. "
                + "Otherwise please try adding this entry manually until more APIs are "
                + "available in coBib."
            )
            LOGGER.warning(msg)
            return OrderedDict()
        label = ""
        entry = {}
        for key, value in contents[list(contents.keys())[0]].items():
            if key in ["title", "url"]:
                entry[key] = value
            elif key == "number_of_pages":
                # we explicitly convert to a string to prevent type errors in the bibtexparser
                str_val = str(value)
                entry["pages"] = int(str_val) if str_val.isnumeric() else str_val
            elif key == "publish_date":
                entry["date"] = value
                try:
                    match = re.search(r"\d{4}", value)
                    if match is None:
                        raise AttributeError  # pragma: no cover
                    entry["year"] = int(match.group())
                    label += str(entry["year"])
                except AttributeError:  # pragma: no cover
                    pass  # pragma: no cover
            elif key == "authors":
                label = value[0]["name"].split()[-1] + label
                entry["author"] = " and ".join([a["name"] for a in value])
            elif key == "publishers":
                entry["publisher"] = " and ".join([a["name"] for a in value])
        # set entry-type do 'book'
        entry["ENTRYTYPE"] = "book"
        bib = OrderedDict()
        bib[label] = Entry(label, entry)

        Event.PostISBNParse.fire(bib)

        return bib

    def dump(self, entry: Entry) -> None:
        """We cannot dump a generic entry as an ISBN."""
        LOGGER.error("Cannot dump an entry as an ISBN.")



"""coBib's selection filter.

This is a simple [`LineFilter`](https://textual.textualize.io/api/filter/#textual.filter.LineFilter)
used to visually style the interactive selection.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from rich.segment import Segment
from rich.style import Style
from textual.color import Color
from textual.filter import LineFilter
from typing_extensions import override


class SelectionFilter(LineFilter):
    """coBib's selection filter."""

    def __init__(self) -> None:
        """Initializes the filter."""
        self.active: bool = False
        """Indicates whether the filter is active, i.e. the selection is not empty."""
        self.selection: set[str] = set()
        """The set of selected labels."""
        self.selection_style = Style(color="white", bgcolor="magenta")
        """The style applied to selected labels."""

    @override
    def apply(self, segments: list[Segment], background: Color | None = None) -> list[Segment]:
        # pylint: disable=unused-argument
        return [
            Segment(
                text,
                self.selection_style
                if any(sel == text.strip() for sel in self.selection)
                else style,
                None,
            )
            for text, style, _ in segments
        ]
